'''
Classes here which inherit from django.db.models.Model define the structure of the database

Django database modules. See https://docs.djangoproject.com/en/dev/intro/tutorial01/
for a basic introduction
'''

import os
os.environ['DJANGO_SETTINGS_MODULE'] = 'db.settings'
import json
import pickle
import inspect
import numpy as np
import importlib
import subprocess
import traceback
import imp
import tables
import tempfile
import shutil
from collections import OrderedDict
from django.db import models
from django.core.exceptions import ObjectDoesNotExist

from riglib import calibrations, experiment

def import_by_path(import_path):
    path_components = import_path.split(".")
    module_name = (".").join(path_components[:-1])
    class_name = path_components[-1]
    module = importlib.import_module(module_name)
    cls = getattr(module, class_name)
    return cls

def func_or_class_to_json(func_or_class, current_values, desc_lookup):
    '''
    Helper for translating a function or class into json parameters for the UI
    '''
    try:
        args = inspect.getargspec(func_or_class)
        names, defaults = args.args, args.defaults
        if "self" in names: names.remove("self")
    except TypeError:
        args = inspect.getargspec(func_or_class.__init__)
        names, defaults = args.args, args.defaults
        names.remove("self")

    params = OrderedDict()
    if len(names) == 0:
        return params
        
    for name, default in zip(names, defaults):
        if name == 'exp':
            continue
        if type(default) is tuple:
            typename = "Tuple"
        elif type(default) is int or type(default) is float:
            typename = "Float"
        elif type(default) is bool:
            typename = "Bool"
        else:
            typename = "String"

        params[name] = dict(type=typename, default=default, desc=desc_lookup(name))
        if name in current_values:
            params[name]['value'] = current_values[name]

    return params

class Task(models.Model):
    name = models.CharField(max_length=128)
    visible = models.BooleanField(default=True, blank=True)
    import_path = models.CharField(max_length=200, blank=True, null=True)

    def __str__(self):
        if not self.import_path is None and self.import_path != "":
            return "Task[{}]: {}".format(self.name, self.import_path)
        else:
            return "Task[{}]".format(self.name)

    def __repr__(self):
        return self.__str__()

    def get_base_class(self):
        if not self.import_path is None and len(self.import_path) > 0:
            return import_by_path(self.import_path)
        else:
            #print(r"Could not find base class for task %s. No import_path provided" % self.name)
            return experiment.Experiment

    def get(self, feats=(), verbose=False):
        if verbose: print("models.Task.get()")

        feature_classes = Feature.getall(feats)
        task_cls = self.get_base_class()

        if not None in feature_classes:
            try:
                return experiment.make(task_cls, feature_classes)
            except:
                print("Problem making the task class!")
                traceback.print_exc()
                print(self.name)
                print(feats)
                print(Feature.getall(feats))
                print("*******")
                return experiment.Experiment
        else:
            print("Task was not installed properly, defaulting to generic experiment!")
            return experiment.Experiment

    @staticmethod
    def add_new_task(task_name, class_path):
        """ Add new task to the database

        Parameters
        ----------
        task_name : string
            Human-readable name of task.
        class_path : string
            Python import path to the class, for example, riglib.experiment.Experiment
        """
        if " " in class_path:
            raise ValueError("Class path cannot have spaces!")

        try:
            path_components = class_path.split(".")
            module_name = (".").join(path_components[:-1])
            class_name = path_components[-1]
            module = importlib.import_module(module_name)
            task_cls = module.getattr(class_name)
        except:
            raise ImportError("Error importing class")
            traceback.print_exc()
        else:
            # finish adding to table
            Task(name=task_name, import_path=class_path).save()

    def params(self, feats=(), values=None):
        '''
        Get user-editable parameters for the frontend

        Parameters
        ----------
        feats : iterable of Feature instances
            Features selected on the task interface
        values : dict
            Values for the task parameters

        '''
        Exp = self.get(feats=feats)
        params = Exp.get_params()

        if values is None:
            values = dict()

        for trait_name in params:
            if params[trait_name]['type'] in ['InstanceFromDB', 'DataFile']:
                mdl_name, filter_kwargs = params[trait_name]['options']

                # get the database Model class from 'db.tracker.models'
                Model = globals()[mdl_name]

                # look up database records which match the model type & filter parameters
                insts = Model.objects.filter(**filter_kwargs).order_by("-date")
                params[trait_name]['options'] = [(i.pk, i.path) for i in insts]

            if trait_name in values:
                params[trait_name]['value'] = values[trait_name]
        return params
        # #from namelist import instance_to_model, instance_to_model_filter_kwargs


        # # Use an ordered dict so that params actually stay in the order they're added, instead of random (hash) order
        # params = OrderedDict()

        # # Run the meta-class constructor to make the Task class (base task class + features )
        # Exp = self.get(feats=feats)
        # ctraits = Exp.class_traits()

        # def add_trait(trait_name):
        #     trait_params = dict()
        #     trait_params['type'] = ctraits[trait_name].trait_type.__class__.__name__
        #     trait_params['default'] = _get_trait_default(ctraits[trait_name])
        #     trait_params['desc'] = ctraits[trait_name].desc
        #     trait_params['hidden'] = 'hidden' if Exp.is_hidden(trait_name) else 'visible'
        #     if hasattr(ctraits[trait_name], 'label'):
        #         trait_params['label'] = ctraits[trait_name].label
        #     else:
        #         trait_params['label'] = trait_name

        #     if trait_name in values:
        #         trait_params['value'] = values[trait_name]

        #     if trait_params['type'] == "InstanceFromDB":
        #         # look up the model name in the trait
        #         mdl_name = ctraits[trait_name].bmi3d_db_model

        #         # get the database Model class from 'db.tracker.models'
        #         Model = globals()[mdl_name]
        #         filter_kwargs = ctraits[trait_name].bmi3d_query_kwargs

        #         # look up database records which match the model type & filter parameters
        #         insts = Model.objects.filter(**filter_kwargs).order_by("-date")
        #         trait_params['options'] = [(i.pk, i.path) for i in insts]

        #     elif trait_params['type'] == 'Instance':
        #         raise ValueError("You should use the 'InstanceFromDB' trait instead of the 'Instance' trait!")

        #     # if the trait is an enumeration, look in the 'Exp' class for
        #     # the options because for some reason the trait itself can't
        #     # store the available options (at least at the time this was written..)
        #     elif trait_params['type'] == "Enum":
        #         raise ValueError("You should use the 'OptionsList' trait instead of the 'Enum' trait!")

        #     elif trait_params['type'] == "OptionsList":
        #         trait_params['options'] = ctraits[trait_name].bmi3d_input_options

        #     elif trait_params['type'] == "DataFile":
        #         # look up database records which match the model type & filter parameters
        #         filter_kwargs = ctraits[trait_name].bmi3d_query_kwargs
        #         insts = DataFile.objects.filter(**filter_kwargs).order_by("-date")
        #         trait_params['options'] = [(i.pk, i.path) for i in insts]

        #     params[trait_name] = trait_params

        #     if trait_name == 'bmi': # a hack for really old data, where the 'decoder' was mistakenly labeled 'bmi'
        #         params['decoder'] = trait_params

        # # add all the traits that are explicitly instructed to appear at the top of the menu
        # ordered_traits = Exp.ordered_traits
        # for trait in ordered_traits:
        #     if trait in Exp.class_editable_traits():
        #         add_trait(trait)

        # # add all the remaining non-hidden traits
        # for trait in Exp.class_editable_traits():
        #     if trait not in params and not Exp.is_hidden(trait):
        #         add_trait(trait)

        # # add any hidden traits
        # for trait in Exp.class_editable_traits():
        #     if trait not in params:
        #         add_trait(trait)

    def sequences(self):
        from .json_param import Parameters
        seqs = dict()
        for s in Sequence.objects.filter(task=self.id):
            try:
                seqs[s.id] = s.to_json()
            except:
                print('Sequence cannot be Accessed: ', s.id)
        return seqs

    def get_generators(self):
        # Supply sequence generators which are declared to be compatible with the selected task class
        exp_generators = []
        Exp = self.get()
        if hasattr(Exp, 'sequence_generators'):
            for seqgen_name in Exp.sequence_generators:
                try:
                    g = Generator.objects.using(self._state.db).get(name=seqgen_name)
                    exp_generators.append([g.id, seqgen_name])
                except:
                    print("missing generator %s" % seqgen_name)
                    traceback.print_exc()
        return exp_generators

    def controls(self, feats=()):
        exp = self.get(feats=feats)
        ctl = exp.controls
        values = dict()

        def desc_lookup(name):
            table = {
                'msg': 'Metadata to mark this moment in experiment',
            }
            if name in table.keys():
                return table[name]
            else:
                return name

        controls = []
        for c in ctl:
            params = func_or_class_to_json(c, values, desc_lookup)
            if 'static' in params:
                params.pop('static')
                controls.append(dict(name=c.__name__, params=params, static=True))
            else:
                controls.append(dict(name=c.__name__, params=params))

        return controls

def can_be_int(x):
    try:
        int(x)
        return True
    except ValueError:
        return False

class Feature(models.Model):
    name = models.CharField(max_length=128)
    visible = models.BooleanField(blank=True, default=True)
    import_path = models.CharField(max_length=200, blank=True, null=True)

    def __str__(self):
        return "Feature[{}]".format(self.name)
    def __repr__(self):
        return self.__str__()

    @property
    def desc(self):
        feature_cls = self.get()
        if not feature_cls is None:
            return feature_cls.__doc__
        else:
            return ''

    def get(self, update_builtin=False):
        from features import built_in_features
        if self.import_path is not None:
            return import_by_path(self.import_path)
        elif self.name in built_in_features:
            import_path = built_in_features[self.name].__module__ + '.' + built_in_features[self.name].__qualname__
            if update_builtin:
                self.import_path = import_path
                self.save()
                return import_by_path(self.import_path)
            else:
                #print("Feature %s import path not found, but found a default" % self.name)
                return import_by_path(import_path)
        else:
            print("Feature %s has no import_path" % self.name)
            return None

    @staticmethod
    def getall(feats):
        feature_class_list = []
        for feat in feats:
            if isinstance(feat, Feature):
                feat_cls = feat.get()
            elif isinstance(feat, int):
                feat_cls = Feature.objects.get(pk=feat).get()
            elif isinstance(feat, str) and can_be_int(feat):
                feat_cls = Feature.objects.get(pk=int(feat)).get()
            elif isinstance(feat, str):
                feat_cls = Feature.objects.get(name=feat).get()
            else:
                print("Cannot find feature: ", feat)

            feature_class_list.append(feat_cls)
        return feature_class_list

class System(models.Model):
    """ Representation for systems which generate data (similar to a data type)"""
    name = models.CharField(max_length=128)
    path = models.TextField()
    archive = models.TextField()
    processor_path = models.CharField(max_length=200, blank=True, null=True)
    input_path = models.TextField(blank=True, null=True)

    def __str__(self):
        return "System[{}]".format(self.name)
    def __repr__(self):
        return self.__str__()

    @staticmethod
    def populate():
        for name in ["eyetracker", "hdf", "plexon", "bmi", "bmi_params", "juice_log", "blackrock"]:
            try:
                System.objects.get(name=name)
            except ObjectDoesNotExist:
                System(name=name, path="/storage/rawdata/%s"%name).save()

    def get_post_processor(self):
        if self.processor_path is not None and len(self.processor_path) > 0:
            return import_by_path(self.processor_path)
        else:
            return lambda x: x # identity fn

    @staticmethod
    def make_new_sys(name):
        try:
            new_sys_rec = System.objects.get(name=name)
        except ObjectDoesNotExist:
            data_dir = "/storage/rawdata/%s" % name
            new_sys_rec = System(name=name, path=data_dir)
            new_sys_rec.save()
            os.popen('mkdir -p %s' % data_dir)

        return new_sys_rec

    def save_to_file(self, obj, filename, obj_name=None, entry_id=-1):
        full_filename = os.path.join(self.path, filename)
        pickle.dump(obj, open(full_filename, 'w'))

        if obj_name is None:
            obj_name = filename.rstrip('.pkl')

        df = DataFile()
        df.path = filename
        df.system = self
        df.entry_id = entry_id
        df.save()

class Subject(models.Model):
    name = models.CharField(max_length=128)
    def __str__(self):
        return "Subject[{}]".format(self.name)
    def __repr__(self):
        return self.__str__()

    @staticmethod
    def get_all_subjects(dbname=None):
        subjects = Subject.objects.using(dbname).all().order_by("name")
        return [s.name for s in subjects]

class Experimenter(models.Model):
    name = models.CharField(max_length=128)
    def __str__(self):
        return "Experimenter[{}]".format(self.name)
    def __repr__(self):
        return self.__str__()

    @staticmethod
    def get_all_experimenters(dbname=None):
        experimenters = Experimenter.objects.using(dbname).all().order_by("name")
        return [s.name for s in experimenters]

class Generator(models.Model):
    name = models.CharField(max_length=128)
    params = models.TextField()
    static = models.BooleanField()
    visible = models.BooleanField(blank=True, default=True)

    def __str__(self):
        return "Generator[{}]".format(self.name)

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def get_all_generators(dbname=None):
        generator_names = []
        generator_functions = []
        tasks = Task.objects.using(dbname).all()
        for task in tasks:
            try:
                task_cls = task.get()
            except:
                # if a task is not importable, then it cannot have any detectable generators
                continue

            if hasattr(task_cls, 'sequence_generators'):
                generator_function_names = task_cls.sequence_generators
                gen_fns = [getattr(task_cls, x) if hasattr(task_cls, x) else None for x in generator_function_names]
                for fn_name, fn in zip(generator_function_names, gen_fns):
                    if fn in generator_functions or fn is None:
                        pass
                    else:
                        generator_names.append(fn_name)
                        generator_functions.append(fn)

        generators = dict()
        for fn_name, fn in zip(generator_names, generator_functions):
            generators[fn_name] = fn
        return generators

    def get(self):
        '''
        Retrieve the function that can be used to construct the ..... generator? sequence?
        '''
        generators = Generator.get_all_generators()
        return generators[self.name]

    @staticmethod
    def remove_unused():
        generators = Generator.get_all_generators()
        listed_generators = set(generators.keys())
        db_generators = set(gen.name for gen in Generator.objects.all())

        # determine which generators are unused in the database using set subtraction
        unused_generators = db_generators - listed_generators
        for name in unused_generators:
            try:
                Generator.objects.filter(name=name).delete()
            except models.ProtectedError:
                pass

    @staticmethod
    def populate():
        generators = Generator.get_all_generators()
        listed_generators = set(generators.keys())
        db_generators = set(gen.name for gen in Generator.objects.all())

        # determine which generators are missing from the database using set subtraction
        missing_generators = listed_generators - db_generators
        for name in missing_generators:

            # TODO not sure why we're populating the 'params' field here since it never gets used    

            # The sequence/generator constructor can either be a callable or a class constructor... not aware of any uses of the class constructor
            try:
                args = inspect.getargspec(generators[name]).args
                print(args)
            except TypeError:
                args = inspect.getargspec(generators[name].__init__).args
                args.remove("self")

            # A generator is determined to be static only if it takes an "exp" argument representing the Experiment class
            static = not ("exp" in args)
            if "exp" in args:
                args.remove("exp")

            # TODO not sure why the 'length' argument is being removed; is it assumed that all generators will take a 'length' argument?
            # if "length" in args:
            #     args.remove("length")

            gen_obj = Generator(name=name, params=",".join(args), static=static)
            gen_obj.save()

    def to_json(self, values=None):
        if values is None:
            values = dict()
        gen = self.get()

        def desc_lookup(name):
            table = {
                'nblocks': 'Number of trials times number of unique targets',
                'ntrials': 'Number of trials',
                'nreps': 'The number of repetitions of each unique condition.',
                'ntargets': 'Number of (evenly spaced) targets',
                'pos': 'Position of the target',
                'distance': 'The distance in cm between the center and peripheral targets',
                'origin': 'Location of the central targets around which the peripheral targets span',
                'boundaries': 'The limits of the allowed target locations',
                'chain_length': 'Number of targets in each sequence before a reward',
                'target_idx': 'Select target indices to include in the task',
            }
            if name in table.keys():
                return table[name]
            else:
                return name

        params = func_or_class_to_json(gen, values, desc_lookup)
        return dict(name=self.name, params=params)

class Sequence(models.Model):
    date = models.DateTimeField(auto_now_add=True)
    generator = models.ForeignKey(Generator, on_delete=models.PROTECT)
    name = models.CharField(max_length=256)
    params = models.TextField() #json data
    sequence = models.TextField(blank=True) #pickle data
    task = models.ForeignKey(Task, on_delete=models.PROTECT)

    def __str__(self):
        return "Sequence[{}] of type Generator[{}]".format(self.name, self.generator.name)

    def __repr__(self):
        return self.__str__()

    def get(self):
        from riglib.experiment import generate
        from .json_param import Parameters

        if hasattr(self, 'generator') and self.generator.static: # If the generator is static, (NOTE: the generator being static is different from the *sequence* being static)
            if len(self.sequence) > 0:
                import ast
                return generate.runseq, dict(seq=pickle.loads(ast.literal_eval(self.sequence)))
            else:
                return generate.runseq, dict(seq=self.generator.get()(**Parameters(self.params).params))
        else:
            return self.generator.get(), Parameters(self.params).params

    def to_json(self):
        from .json_param import Parameters
        state = 'saved' if self.pk is not None else "new"
        js = dict(name=self.name, state=state)
        js['static'] = len(self.sequence) > 0
        js['params'] = self.generator.to_json(Parameters(self.params).params)['params']
        js['generator'] = self.generator.id, self.generator.name
        return js

    @classmethod
    def from_json(cls, js):
        '''
        Construct a models.Sequence instance from JSON data (e.g., generated by the web interface for starting experiments)
        '''
        from .json_param import Parameters

        # Error handling when input argument 'js' actually specifies the primary key of a Sequence object already in the database
        try:
            seq = Sequence.objects.get(pk=int(js))
            print("retreiving sequence from POSTed ID")
            return seq
        except:
            pass

        # Make sure 'js' is a python dictionary
        if not isinstance(js, dict):
            js = json.loads(js)

        # Determine the ID of the "generator" used to make this sequence
        genid = js['generator']
        if isinstance(genid, (tuple, list)):
            genid = genid[0]

        # Construct the database record for the new Sequence object
        seq = cls(generator_id=int(genid), name=js['name'])

        # Link the generator instantiation parameters to the sequence record
        # Parameters are stored in JSON format in the database
        seq.params = Parameters.from_html(js['params']).to_json()

        # If the sequence is to be static,
        if js['static']:
            print("db.tracker.models.Sequence.from_json: storing static sequence data to database")
            generator_params = Parameters(seq.params).params
            seq_data = seq.generator.get()(**generator_params)
            seq.sequence = pickle.dumps(seq_data)
        return seq

class TaskEntry(models.Model):
    subject = models.ForeignKey(Subject, on_delete=models.PROTECT)
    experimenter = models.ForeignKey(Experimenter, null=True, on_delete=models.PROTECT)
    date = models.DateTimeField(auto_now_add=True)
    task = models.ForeignKey(Task, on_delete=models.PROTECT)
    feats = models.ManyToManyField(Feature)
    sequence = models.ForeignKey(Sequence, blank=True, null=True, on_delete=models.PROTECT)
    project = models.TextField()
    session = models.TextField()
    rig_name = models.TextField(default="pagaiisland2")

    params = models.TextField()
    report = models.TextField()
    notes = models.TextField()
    metadata = models.TextField(default="") # serialized custom key/value metadata
    visible = models.BooleanField(blank=True, default=True)
    backup = models.BooleanField(blank=True, default=False)
    template = models.BooleanField(blank=True, default=False)
    entry_name = models.CharField(blank=True, null=True, max_length=50)
    sw_version = models.CharField(blank=True, null=True, max_length=100)

    def __str__(self):
        return "{date}: {subj} on {task} task, id={id}".format(
            date=str(self.date),
            subj=self.subject.name,
            task=self.task.name,
            id=self.id)

    def __repr__(self):
        return self.__str__()

    @property
    def ui_id(self):
        if self.entry_name is not None and self.entry_name != "":
            return "%s (%d)" % (self.entry_name, self.id)
        else:
            return str(self.id)

    def get(self, feats=()):
        from .json_param import Parameters
        from riglib import experiment
        Exp = experiment.make(self.task.get(), tuple(f.get() for f in self.feats.all())+feats)
        params = Parameters(self.params)
        params.trait_norm(Exp.class_traits())
        if issubclass(Exp, experiment.Sequence):
            gen, gp = self.sequence.get()
            seq = gen(Exp, **gp)
            exp = Exp(seq, **params.params)
        else:
            exp = Exp(**params.params)
        exp.event_log = json.loads(self.report)
        return exp

    def update_date(self, *date_args):
        """Utility function to update the date of a record if set improperly (e.g., if entered manually)"""
        import datetime
        if len(date_args) == 1 and isinstance(date_args[0], datetime.datetime):
            self.date = date_args[0]
            self.save()
        elif len(date_args) >= 3 and len(date_args) <= 7: # assume these are numbers with fields split out
            self.date = datetime.datetime(*date_args)
            self.save()
        else:
            raise ValueError("Unrecognized date arguments: ", date_args)

    @property
    def task_params(self):
        from .json_param import Parameters
        data = Parameters(self.params).params

        if 'bmi' in data:
            data['decoder'] = data['bmi']
        ##    del data['bmi']
        return data

    @property
    def task_metadata(self):
        from .json_param import Parameters
        data = Parameters(self.metadata).params
        return data

    @property
    def sequence_params(self):
        from .json_param import Parameters
        data = Parameters(self.sequence.params).params
        return data

    @staticmethod
    def get_default_metadata():
        subject = {
            'type': 'Enum',
            'default': '',
            'desc': 'Who',
            'hidden': 'visible',
            'options':  Subject.get_all_subjects(),
            'required': True
        }
        experimenter = {
            'type': 'Enum',
            'default': '',
            'desc': 'Who is running the experiment',
            'hidden': 'visible',
            'options':  Experimenter.get_all_experimenters(),
            'required': True
        }
        project = {
            'type': 'String',
            'default': '',
            'desc': 'Which project are you working on',
            'hidden': 'visible',
            'value': '',
            'required': True
        }
        session = {
            'type': 'String',
            'default': '',
            'desc': 'Specific instance of the project',
            'hidden': 'visible',
            'value': '',
            'required': True
        }
        metadata = {
            'subject': subject,
            'experimenter': experimenter,
            'project': project,
            'session': session,
        }
        return metadata

    def plexfile(self, path='/storage/plexon/', search=False):
        rplex = Feature.objects.get(name='relay_plexon')
        rplexb = Feature.objects.get(name='relay_plexbyte')
        feats = self.feats.all()
        if rplex not in feats and rplexb not in feats:
            return None

        if not search:
            system = System.objects.get(name='plexon')
            df = DataFile.objects.filter(entry=self.id, system=system)
            if len(df) > 0:
                return df[0].get_path()

        if len(self.report) > 0:
            event_log = json.loads(self.report)
            import os, sys, glob, time
            if len(event_log) < 1:
                return None

            start = event_log[-1][2]
            files = sorted(glob.glob(path+"/*.plx"), key=lambda f: abs(os.stat(f).st_mtime - start))

            if len(files) > 0:
                tdiff = os.stat(files[0]).st_mtime - start
                if abs(tdiff) < 60:
                     return files[0]

    def offline_report(self):
        Exp = self.task.get(self.feats.all())

        if len(self.report) == 0:
            return dict()
        else:
            report = json.loads(self.report)
            if isinstance(report, list):
                # old method: calculate from full event log
                rpt = Exp.offline_report(report)
            else:
                # new method: reformat stats
                rpt = Exp.format_log_summary(report)


            ## If this is a BMI block, add the decoder name to the report (doesn't show up properly in drop-down menu for old blocks)
            # try:
            #     from db import dbfunctions
            #     te = dbfunctions.TaskEntry(self.id, dbname=self._state.db)
            #     rpt['Decoder name'] = te.decoder_record.name + ' (trained in block %d)' % te.decoder_record.entry_id
            # except AttributeError:
            #     pass
            # except:
            #     import traceback
            #     traceback.print_exc()
            return rpt

    def to_json(self):
        '''
        Create a JSON dictionary of the metadata associated with this block for display in the web interface
        '''
        print("starting TaskEntry.to_json()")
        from .json_param import Parameters

        # Run the metaclass constructor for the experiment used. If this can be avoided, it would help to break some of the cross-package software dependencies,
        # making it easier to analyze data without installing software for the entire rig

        Exp = self.task.get(self.feats.all())
        from . import exp_tracker
        tracker = exp_tracker.get()
        if tracker.get_status() and tracker.proc is not None and hasattr(tracker.proc, 'saveid') and tracker.proc.saveid == self.id:
            state = tracker.get_status()#'completed' if self.pk is not None else "new"
        else:
            state = 'completed'  
        js = dict(task=self.task.id, state=state, subject=self.subject.id, notes=self.notes)
        js['feats'] = dict([(f.id, f.name) for f in self.feats.all()])
        js['params'] = self.task.params(self.feats.all(), values=self.task_params)
        js['controls'] = self.task.controls(feats=self.feats.all())

        if len(js['params'])!=len(self.task_params):
            print('param lengths: JS:', len(js['params']), 'Task: ', len(self.task_params))

        # Add metadata
        js['metadata'] = self.get_default_metadata()
        js['metadata']['subject']['default'] = self.subject.name
        if self.experimenter: 
            js['metadata']['experimenter']['default'] = self.experimenter.name
        js['metadata']['project']['value'] = self.project
        js['metadata']['session']['value'] = self.session
        js['metadata'].update(dict([
            (
                name, 
                {
                    'type': 'String',
                    'default': '',
                    'desc': '',
                    'hidden': 'visible',
                    'value': value,
                    'required': False
                }
            ) for name, value in self.task_metadata.items()
        ]))
        

        # Supply sequence generators which are declared to be compatible with the selected task class
        exp_generators = dict()
        if hasattr(Exp, 'sequence_generators'):
            for seqgen_name in Exp.sequence_generators:
                try:
                    g = Generator.objects.using(self._state.db).get(name=seqgen_name)
                    exp_generators[g.id] = seqgen_name
                except:
                    print("missing generator %s" % seqgen_name)
        js['generators'] = exp_generators

        ## Add the sequence, used when the block gets copied
        print("getting the sequence, if any")
        if issubclass(self.task.get(), experiment.Sequence):
            js['sequence'] = {self.sequence.id:self.sequence.to_json()}

        datafiles = DataFile.objects.using(self._state.db).filter(entry=self.id)

        ## Add data files linked to this task entry to the web interface.
        backup_root = KeyValueStore.get('backup_root', '/None')

        js['datafiles'] = dict()
        system_names = set(d.system.name for d in datafiles)
        for name in system_names:
            js['datafiles'][name] = [d.get_path() + ' (backup status: %s)' % d.backup_status for d in datafiles if d.system.name == name]

        # Parse the "report" data and put it into the JS response
        js['report'] = self.offline_report()

        recording_sys_make = KeyValueStore.get('recording_sys')

        _neuralinfo = dict(is_seed=Exp.is_bmi_seed, length=0, name='', units=[])
        js['bmi'] = dict(_neuralinfo=_neuralinfo)
        if recording_sys_make == 'plexon':
            try:
                from plexon import plexfile # keep this import here so that only plexon rigs need the plexfile module installed
                plexon = System.objects.using(self._state.db).get(name='plexon')
                df = DataFile.objects.using(self._state.db).get(entry=self.id, system=plexon)

                _neuralinfo = dict(is_seed=Exp.is_bmi_seed)
                if Exp.is_bmi_seed:
                    plx = plexfile.openFile(df.get_path().encode('utf-8'), load=False)
                    path, name = os.path.split(df.get_path())
                    name, ext = os.path.splitext(name)

                    _neuralinfo['length'] = plx.length
                    _neuralinfo['units'] = plx.units
                    _neuralinfo['name'] = name

                js['bmi'] = dict(_neuralinfo=_neuralinfo)
            except MemoryError:
                print("Memory error opening plexon file!")
                js['bmi'] = dict(_neuralinfo=None)
            except (ObjectDoesNotExist, AssertionError, IOError):
                print("No plexon file found")
                js['bmi'] = dict(_neuralinfo=None)

        elif recording_sys_make == 'blackrock':
            try:
                print('skipping .nev conversion')
                js['bmi'] = dict(_neuralinfo=None)

                # length, units = parse_blackrock_file(self.nev_file, self.nsx_files, self)

                # js['bmi'] = dict(_neuralinfo=dict(
                #     length=length,
                #     units=units,
                #     name=name,
                #     is_seed=int(Exp.is_bmi_seed),
                #     ))

            except (ObjectDoesNotExist, AssertionError, IOError):
                print("No blackrock files found")
                js['bmi'] = dict(_neuralinfo=None)
            except:
                import traceback
                traceback.print_exc()
                js['bmi'] = dict(_neuralinfo=None)
        elif recording_sys_make == 'TDT':
            print('This code does not yet know how to open TDT files!')
            js['bmi'] = dict(_neuralinfo=None)
            #raise NotImplementedError("This code does not yet know how to open TDT files!")
        elif recording_sys_make == 'ecube':
            try:
                sys = System.objects.using(self._state.db).get(name='ecube')
                df = DataFile.objects.using(self._state.db).get(entry=self.id, system=sys)
                filepath = df.get_path()

                from riglib.ecube import parse_file
                _neuralinfo = dict(is_seed=Exp.is_bmi_seed)
                if Exp.is_bmi_seed:
                    try:
                        info = parse_file(str(df.get_path()))
                        path, name = os.path.split(df.get_path())
                        name, ext = os.path.splitext(name)

                        _neuralinfo['length'] = info.length
                        _neuralinfo['units'] = info.units
                        _neuralinfo['name'] = name
                    except:
                        _neuralinfo['is_seed'] = False

                js['bmi'] = dict(_neuralinfo=_neuralinfo)
            except ModuleNotFoundError:
                print("ecube reader not installed")
                js['bmi'] = dict(_neuralinfo=None)
            except (ObjectDoesNotExist, IOError):
                print("No ecube file found")
                js['bmi'] = dict(_neuralinfo=None)
        else:
            print('Unrecognized recording_system!')

        for dec in Decoder.objects.using(self._state.db).filter(entry=self.id):
            if 'bmi' not in js:
                print("Warning: found a decoder but not a recording system. Your recording system may not be set up correctly")
                js['bmi'] = dict(_neuralinfo=None)
            js['bmi'][dec.name] = dec.to_json()

        # collections info
        js['collections'] = []
        for col in TaskEntryCollection.objects.all():
            if self in col.entries.all():
                js['collections'].append(col.name)

        js['plot_files'] = dict()  # deprecated
        js['flagged_for_backup'] = self.backup
        js['visible'] = self.visible
        js['template'] = self.template
        entry_name = self.entry_name if not self.entry_name is None else ""
        js['entry_name'] = entry_name
        print("TaskEntry.to_json finished!")
        return js

    @property
    def plx_file(self):
        '''
        Returns the name of the plx file associated with the session.
        '''
        plexon = System.objects.get(name='plexon')
        try:
            df = DataFile.objects.get(system=plexon, entry=self.id)
            return os.path.join(df.system.path, df.path)
        except:
            import traceback
            traceback.print_exc()
            return 'noplxfile'

    @property
    def nev_file(self):
        '''
        Return the name of the nev file associated with the session.
        '''
        try:
            df = DataFile.objects.get(system__name="blackrock", path__endswith=".nev", entry=self.id)
            return df.get_path()
        except:
            try:
                df = DataFile.objects.get(system__name="blackrock2", path__endswith=".nev", entry=self.id)
                return df.get_path()
            except:
                return None
            #import traceback
            #traceback.print_exc()
            #return 'no_nev_file'
            #return None

    @property
    def nsx_files(self):
        '''Return a list containing the names of the nsx files (there could be more
        than one) associated with the session.

        nsx files extensions are .ns1, .ns2, ..., .ns6
        '''
        try:
            dfs = []
            for k in range(1, 7):
                df_k = DataFile.objects.filter(system__name="blackrock", path__endswith=".ns%d" % k, entry=self.id)
                dfs += list(df_k)

            return [df.get_path() for df in dfs]
        except:
            try:
                dfs = []
                for k in range(1, 7):
                    df_k = DataFile.objects.filter(system__name="blackrock2", path__endswith=".ns%d" % k, entry=self.id)
                    dfs+= list(df_k)
                return [df.get_path() for df in dfs]
            except:
                return None

    @property
    def name(self):
        '''
        Return a string representing the 'name' of the block. Note that the block
        does not really have a unique name in the current implementation.
        Thus, the 'name' is a hack this needs to be hacked because the current way of determining a
        a filename depends on the number of things in the database, i.e. if
        after the fact a record is removed, the number might change. read from
        the file instead
        '''
        recording_sys_make = KeyValueStore.get('recording_sys')

        if recording_sys_make == 'plexon':
            try:
                return str(os.path.basename(self.plx_file).rstrip('.plx'))
            except:
                return 'noname'
        elif recording_sys_make == 'blackrock':
            try:
                return str(os.path.basename(self.nev_file).rstrip('.nev'))
            except:
                return 'noname'
        else:
            raise Exception('Unrecognized recording_system!')

    @classmethod
    def from_json(cls, js):
        pass

    def get_decoder(self, dbname=None):
        """
        Get the Decoder instance associated with this task entry
        """
        params = eval(self.params)
        decoder_id = params['bmi']
        return Decoder.objects.using(dbname).get(id=decoder_id)

    @property
    def desc(self):
        """Get a description of the TaskEntry using the experiment class and the
        record-specific parameters """
        from . import json_param
        params = json_param.Parameters(self.params)

        if self.report is not None and len(self.report) > 0:
            try:
                report_data = json.loads(self.report)
            except:
                print(self.id)
                print("THIS TE IS BROKEN <--")
                report_data = None

        else:
            report_data = None
        try:
            Exp = self.task.get_base_class()
            return Exp.get_desc(params.get_data(), report_data)
        except:
            return "Error generating description"

    def get_data_files(self, dbname=None):
        return list(DataFile.objects.using(dbname).filter(entry_id=self.id))

    def get_data_files_dict(self, data_dir="", dbname=None):
        file_list = self.get_data_files(dbname=dbname)
        files = {}
        for df in file_list:
            files[df.system.name] = os.path.join(data_dir, df.system.name, os.path.basename(df.path))
        return files

    def get_data_files_dict_absolute(self, dbname=None):
        file_list = self.get_data_files(dbname=dbname)
        files = {}
        for df in file_list:
            files[df.system.name] = os.path.join(df.system.path, os.path.basename(df.path))
        return files

    def make_hdf_self_contained(self, data_dir=None, dbname=None):
        '''
        If the entry has an hdf file associated with it, dump the entry metadata into it
        '''
        try:
            df = DataFile.objects.using(dbname).get(entry__id=self.id, system__name="hdf")
        except:
            print("No HDF file to make self contained")
            return False

        import h5py
        if data_dir is None:
            h5file = df.get_path()
        else:
            h5file = os.path.join(data_dir, df.system.name, os.path.basename(df.path))
        try:
            hdf = h5py.File(h5file, mode='a')
        except:
            print("Cannot open HDF file")
            return False
        print("Updating hdf file with metadata for task entry %d" % self.id)
        print(h5file)

        # Add any task metadata
        hdf['/'].attrs["task_name"] = self.task.name
        hdf['/'].attrs["features"] = [f.name for f in self.feats.all()]
        hdf['/'].attrs["rig_name"] = self.rig_name
        hdf['/'].attrs["block_number"] = self.id
        hdf['/'].attrs["subject"] = self.subject.name
        hdf['/'].attrs["experimenter"] = self.experimenter.name
        hdf['/'].attrs["date"] = str(self.date)
        hdf['/'].attrs["project"] = self.project
        hdf['/'].attrs["session"] = self.session
        if self.sequence is not None:
            hdf['/'].attrs["sequence"] = self.sequence.name
            hdf['/'].attrs["sequence_params"] = self.sequence.params
            hdf['/'].attrs["generator"] = self.sequence.generator.name
            hdf['/'].attrs["generator_params"] = self.sequence.generator.params
        hdf['/'].attrs["report"] = self.report
        hdf['/'].attrs["notes"] = self.notes
        hdf['/'].attrs["sw_version"] = self.sw_version

        # Link any data files
        data_files = []
        for df in DataFile.objects.using(dbname).filter(entry__id=self.id):
            data_files.append(df.get_path())
        hdf['/'].attrs["data_files"] = data_files
        hdf.close()

        # TODO save decoder parameters to hdf file, if applicable

        return True

class Calibration(models.Model):
    subject = models.ForeignKey(Subject, on_delete=models.PROTECT)
    date = models.DateTimeField(auto_now_add=True)
    name = models.CharField(max_length=128)
    system = models.ForeignKey(System, on_delete=models.PROTECT)

    params = models.TextField()

    def __str__(self):
        return "{date}:{system} calibration for {subj}".format(date=self.date,
            subj=self.subject.name, system=self.system.name)

    def __repr__(self):
        return self.__str__()

    def get(self):
        from .json_param import Parameters
        return getattr(calibrations, self.name)(**Parameters(self.params).params)

class AutoAlignment(models.Model):
    date = models.DateTimeField(auto_now_add=True)
    name = models.TextField()

    def __unicode__(self):
        return "{date}:{name}".format(date=self.date, name=self.name)

    def get(self):
        return calibrations.AutoAlign(self.name)


def decoder_unpickler(mod_name, kls_name):
    if kls_name == 'StateSpaceFourLinkTentacle2D':
        kls_name = 'StateSpaceNLinkPlanarChain'
        mod_name = 'riglib.bmi.state_space_models'

    if kls_name == 'StateSpaceEndptVel':
        kls_name = 'LinearVelocityStateSpace'
        mod_name = 'riglib.bmi.state_space_models'

    if kls_name == 'State':
        mod_name = 'riglib.bmi.state_space_models'
    mod = importlib.import_module(mod_name)
    return getattr(mod, kls_name)


class Decoder(models.Model):
    date = models.DateTimeField(auto_now_add=True)
    name = models.CharField(max_length=128)
    entry = models.ForeignKey(TaskEntry, on_delete=models.PROTECT)
    path = models.TextField()

    def __str__(self):
        return "Decoder[{date}:{name}] trained from {entry}".format(date=self.date, name=self.name, entry=self.entry)

    def __repr__(self):
        return self.__str__()

    def get_data_path(self, db_name=None):
        # data_path = KeyValueStore.get('data_path', '', dbname=db_name)
        # if len(data_path) == 0:
        #     print("Database path not set up correctly!")
        # return data_path
        return System.objects.using(db_name).get(name='bmi').path

    @property
    def filename(self):
        data_path = self.get_data_path()
        return os.path.join(data_path, self.path)

    def load(self, db_name=None, **kwargs):
        data_path = self.get_data_path()
        decoder_fname = os.path.join(data_path, self.path)

        if os.path.exists(decoder_fname):
            try:
                return pickle.load(open(decoder_fname, 'rb'), encoding='latin1')
            # try:
            #     fh = open(decoder_fname, 'r')
            #     unpickler = pickle.Unpickler(fh, **kwargs)
            #     unpickler.find_global = decoder_unpickler
            #     dec = unpickler.load() # object will now contain the new class path reference
            #     fh.close()

            #     dec.name = self.name
            #     return dec
            except:
                import traceback
                traceback.print_exc()
                return None
        else: # file not present!
            print("Decoder file could not be found! %s" % decoder_fname)
            return None

    def get(self):
        return self.load()

    def to_json(self):
        dec = self.get()
        decoder_data = dict(name=self.name, path=self.path)
        if not (dec is None):
            decoder_data['cls'] = dec.__class__.__name__
            if hasattr(dec, 'units'):
                decoder_data['units'] = dec.units
            else:
                decoder_data['units'] = []

            if hasattr(dec, 'binlen'):
                decoder_data['binlen'] = dec.binlen
            else:
                decoder_data['binlen'] = 0

            if hasattr(dec, 'tslice'):
                decoder_data['tslice'] = dec.tslice
            else:
                decoder_data['tslice'] = []

        return decoder_data

def parse_blackrock_file_n2h5(nev_fname, nsx_files):
    '''
    # convert .nev file to hdf file using Blackrock's n2h5 utility (if it doesn't exist already)
    # this code goes through the spike_set for each channel in order to:
    #  1) determine the last timestamp in the file
    #  2) create a list of units that had spikes in this file
    '''
    nev_hdf_fname = nev_fname + '.hdf'

    if not os.path.isfile(nev_hdf_fname):
        subprocess.call(['n2h5', nev_fname, nev_hdf_fname])

    nev_hdf = tables.openFile(nev_hdf_fname, 'r')

    last_ts = 0
    units = []

    #for key in [key for key in nev_hdf.get('channel').keys() if 'channel' in key]:
    chans = nev_hdf.root.channel
    chan_names= chans._v_children
    for key in [key for key in list(chan_names.keys()) if 'channel' in key]:
        chan_tab = nev_hdf.root.channel._f_getChild(key)
        if 'spike_set' in chan_tab:
            spike_set = chan_tab.spike_set
            if spike_set is not None:
                tstamps = spike_set[:]['TimeStamp']
                if len(tstamps) > 0:
                    last_ts = max(last_ts, tstamps[-1])
                else:
                    print('skipping ', key, ': no spikes')

                channel = int(key[-5:])
                for unit_num in np.sort(np.unique(spike_set[:]['Unit'])):
                    units.append((channel, int(unit_num)))
        else:
            print('skipping ', key, ': no spikeset')

    fs = 30000.
    nev_length = last_ts / fs
    nsx_lengths = []

    if nsx_files is not None:
        nsx_fs = dict()
        nsx_fs['.ns1'] = 500
        nsx_fs['.ns2'] = 1000
        nsx_fs['.ns3'] = 2000
        nsx_fs['.ns4'] = 10000
        nsx_fs['.ns5'] = 30000
        nsx_fs['.ns6'] = 30000

        NSP_channels = np.arange(128) + 1

        nsx_lengths = []
        for nsx_fname in nsx_files:

            nsx_hdf_fname = nsx_fname + '.hdf'
            if not os.path.isfile(nsx_hdf_fname):
                # convert .nsx file to hdf file using Blackrock's n2h5 utility
                subprocess.call(['n2h5', nsx_fname, nsx_hdf_fname])

            nsx_hdf = h5py.File(nsx_hdf_fname, 'r')

            for chan in NSP_channels:
                chan_str = str(chan).zfill(5)
                path = 'channel/channel%s/continuous_set' % chan_str
                if nsx_hdf.get(path) is not None:
                    last_ts = len(nsx_hdf.get(path).value)
                    fs = nsx_fs[nsx_fname[-4:]]
                    nsx_lengths.append(last_ts / fs)

                    break

    length = max([nev_length] + nsx_lengths)
    return length, units,

def parse_blackrock_file(nev_fname, nsx_files, task_entry, nsx_chan = np.arange(96) + 1):
    ''' Method to parse blackrock files using new
    brpy from blackrock (with some modifications). Files are
    saved as a ____ file?

    # this code goes through the spike_set for each channel in order to:
    #  1) determine the last timestamp in the file
    #  2) create a list of units that had spikes in this file
    '''
    from riglib.blackrock.brpylib import NevFile, NsxFile

    # First parse the NEV file:
    if nev_fname is not None:
        nev_hdf_fname = nev_fname + '.hdf'
        if task_entry is not None:
            datafiles = DataFile.objects.using(task_entry._state.db).filter(entry_id=task_entry.id)
            files = [d.get_path() for d in datafiles]
        else:
            # Guess where the file shoudl be:
            files = ['/storage/rawdata/blackrock/'+nev_hdf_fname]

        if nev_hdf_fname in files:
            hdf = tables.openFile(nev_hdf_fname)
            n_units = hdf.root.attr[0]['n_units']
            last_ts = hdf.root.attr[0]['last_ts']
            units = hdf.root.attr[0]['units'][:n_units]

        else:
            try:
                nev_file = NevFile(nev_fname)
                spk_data = nev_file.getdata()
            except:
                print('nev file is not available for opening. Try in a few seconds!')
                raise Exception

            # Make HDF file from NEV file #
            last_ts, units, h5file = make_hdf_spks(spk_data, nev_hdf_fname)

            if task_entry is not None:
                from . import dbq
                dbq.save_data(nev_hdf_fname, 'blackrock', task_entry.pk, move=False, local=True, custom_suffix='', dbname=task_entry._state.db)

        fs = 30000.
        nev_length = last_ts / fs
    else:
        units = []
        nev_length = 0

    tmax_cts = 0
    if nsx_files is not None:
        for nsx_fname in nsx_files:
            nsx_hdf_fname = nsx_fname + '.hdf'
            if not os.path.isfile(nsx_hdf_fname):

                # convert .nsx file to hdf file using Blackrock's n2h5 utility
                # subprocess.call(['n2h5', nsx_fname, nsx_hdf_fname])
                nsx_file = NsxFile(nsx_fname)

                # Extract data - note: data will be returned based on *SORTED* elec_ids, see cont_data['elec_ids']
                cont_data = nsx_file.getdata(nsx_chan, 0, 'all', 1)

                # Close the nsx file now that all data is out
                nsx_file.close()

                # Make HDF file:
                tmax_cts = make_hdf_cts(cont_data, nsx_hdf_fname, nsx_file)
                if task_entry is not None:
                    from . import dbq
                    dbq.save_data(nsx_hdf_fname, 'blackrock', task_entry.pk, move=False, local=True, custom_suffix='', dbname=task_entry._state.db)
            else:
                tmax_cts = 0

    length = max([nev_length] + [tmax_cts])
    return length, units

def make_hdf_spks(data, nev_hdf_fname):
    last_ts = 0
    units = []

    #### Open h5file: ####
    tf = tempfile.NamedTemporaryFile(delete=False)
    h5file = tables.openFile(tf.name, mode="w", title='BlackRock Nev Data')
    h5file.createGroup('/', 'channel')

    ### Spike Data First ###
    channels = data['spike_events']['ChannelID']
    base_str = 'channel00000'
    for ic, c in enumerate(channels):
        c_str = base_str[:-1*len(str(c))]+ str(c)
        h5file.createGroup('/channel', c_str)
        tab = h5file.createTable('/channel/'+c_str, 'spike_set', spike_set)

        for i, (ts, u, wv) in enumerate(zip(data['spike_events']['TimeStamps'][ic], data['spike_events']['Classification'][ic], data['spike_events']['Waveforms'][ic])):
            trial = tab.row
            last_ts = np.max([last_ts, ts])
            skip = False
            if u == 'none':
                u = 10
            elif u == 'noise':
                skip = True

            if not skip:
                trial['Unit'] = u
                trial['Wave'] = wv
                trial['TimeStamp'] = ts
                trial.append()

        #Check for non-zero units:
        if len(data['spike_events']['TimeStamps'])>0:
            un = np.unique(data['spike_events']['Classification'][ic])
            for ci in un:
                #ci = 10
                if ci == 'none':
                    # Unsorted
                    units.append((c, 10))
                elif ci == 'noise':
                    pass
                else:
                    # Sorted (units are numbered )
                    units.append((c, int(ci)))

        tab.flush()

    ### Digital Data ###
    try:
        ts = data['dig_events']['TimeStamps']
        val = data['dig_events']['Data']

        for dchan in range(1, 1+len(ts)):
            h5file.createGroup('/channel', 'digital000'+str(dchan))
            dtab = h5file.createTable('/channel/digital000'+str(dchan), 'digital_set', digital_set)

            ts_chan = ts[dchan-1]
            val_chan = val[dchan-1]

            for ii, (tsi, vli) in enumerate(zip(ts_chan, val_chan)):
                trial = dtab.row
                trial['TimeStamp'] = tsi
                trial['Value'] = vli
                trial.append()
            last_ts = np.max([last_ts, tsi])
            dtab.flush()
    except:
        print('no digital info in nev file ')

    # Adding length / unit info:
    tb = h5file.createTable('/', 'attr', mini_attr)
    rw = tb.row
    rw['last_ts'] = last_ts
    U = np.zeros((500, 2))
    n_units = len(units)
    U[:n_units, :] = np.vstack((units))

    rw['units'] = U
    rw['n_units'] = n_units
    rw.append()
    tb.flush()

    h5file.close()
    shutil.copyfile(tf.name, nev_hdf_fname)
    os.remove(tf.name)
    print('successfully made HDF file from NEV file: %s' %nev_hdf_fname)

    un_array = np.vstack((units))
    idx = np.lexsort((un_array[:, 1], un_array[:, 0]))
    units2 = [units[i] for i in idx]

    return last_ts, units2, h5file

def make_hdf_cts(data, nsx_hdf_fname, nsx_file):
    last_ts = []
    channels = []

    tf = tempfile.NamedTemporaryFile(delete=False)
    h5file = tables.openFile(tf.name, mode="w", title='BlackRock Nsx Data')
    h5file.createGroup('/', 'channel')

    channel_ids = data['elec_ids']
    hdr_ids = data['ExtendedHeaderIndices']
    channel_labels = [nsx_file.extended_headers[h]['ElectrodeLabel'] for h in hdr_ids]
    print('saving channel nums: ', channel_ids)

    for ic, (c, chan) in enumerate(zip(channel_ids, channel_labels)):
        channel_data = data['data'][ic]
        c_str = 'channel'+str(c).zfill(5)
        h5file.createGroup('/channel', c_str)
        h5file.createArray('/channel/'+c_str, 'Value', channel_data)

    t = data['start_time_s'] + np.arange(data['data'].shape[1]) / data['samp_per_s']
    h5file.createArray('/channel', 'TimeStamp', t)
    #h5file.createArray('/channel', 'ElectrodeLabels', np.hstack((channel_labels)))
    h5file.close()
    shutil.copyfile(tf.name, nsx_hdf_fname)
    os.remove(tf.name)
    print('successfully made HDF file from NSX file: %s' %nsx_hdf_fname)
    return t[-1]

class spike_set(tables.IsDescription):
    TimeStamp = tables.Int32Col()
    Unit = tables.Int8Col()
    Wave = tables.Int32Col(shape=(48,))

class digital_set(tables.IsDescription):
    TimeStamp = tables.Int32Col()
    Value = tables.Int16Col()

class continuous_set(tables.IsDescription):
    TimeStamp = tables.Int32Col()
    Value = tables.Int16Col()

class mini_attr(tables.IsDescription):
    last_ts = tables.Float64Col()
    units = tables.Int16Col(shape=(500, 2))
    n_units = tables.Int16Col()


class DataFile(models.Model):
    date = models.DateTimeField(auto_now_add=True)
    local = models.BooleanField(default=True)
    archived = models.BooleanField(default=False)
    path = models.CharField(max_length=256)
    system = models.ForeignKey(System, on_delete=models.PROTECT, blank=True, null=True)
    entry = models.ForeignKey(TaskEntry, on_delete=models.PROTECT, blank=True, null=True)
    backup_status = models.CharField(max_length=256, blank=True, null=True)

    @staticmethod
    def create(system, task_entry, path, **kwargs):
        df = DataFile(system_id=system.id, entry_id=task_entry.id)

        for attr in kwargs:
            setattr(df, attr, kwargs[attr])

        post_processor = system.get_post_processor()
        df.path = post_processor(path)
        df.save()
        return df

    def __str__(self):
        if self.entry_id > 0:
            return "{name} datafile for {entry}".format(name=self.system.name, entry=self.entry)
        else:
            return "datafile '{name}' for System {sys_name}".format(name=self.path, sys_name=self.system.name)

    def __repr__(self):
        return self.__str__()

    def to_json(self):
        return dict(system=self.system.name, path=self.path)

    def get(self):
        '''
        Open the datafile, if it's of a known type
        '''
        if self.system.name == 'hdf':
            import tables
            return tables.open_file(self.get_path())
        elif self.path[-4:] == '.pkl': # pickle file
            import pickle
            return pickle.load(open(self.get_path()))
        else:
            raise ValueError("models.DataFile does not know how to open this type of file: %s" % self.path)

    def get_path(self, check_archive=False):
        '''
        Get the full path to the file
        '''
        if not check_archive and not self.archived:
            return os.path.join(self.system.path, self.path)

        paths = self.system.archive.split()
        for path in paths:
            fname = os.path.join(path, self.path)
            if os.path.isfile(fname):
                return fname

        raise IOError('File has been lost! '+fname)

    def has_cache(self):
        if self.system.name != "plexon":
            return False

        path, fname = os.path.split(self.get_path())
        fname, ext = os.path.splitext(fname)
        cache = os.path.join(path, '.%s.cache'%fname)
        return os.path.exists(cache)

    def remove(self, **kwargs):
        try:
            os.unlink(self.get_path())
        except OSError:
            print("already deleted...")

    def delete(self, **kwargs):
        self.remove()
        super(DataFile, self).delete(**kwargs)

    def is_backed_up(self, backup_root):
        '''
        Return a boolean indicating whether a copy of the file is available on the backup
        '''
        try:
            fname = self.get_path()
            rel_datafile = os.path.relpath(fname, '/storage')
            backup_fname = os.path.join(backup_root, rel_datafile)
            return os.path.exists(backup_fname)
        except:
            return False

    @property
    def file_size(self):
        try:
            path = self.get_path()
            return os.stat(path).st_size
        except:
            print("Error getting data file size: ", self)
            traceback.print_exc()
            return -1

    def upload_to_cloud(self):
        """Upload file to google cloud storage"""
        full_filename = self.get_path()
        data = dict(full_filename=full_filename, filename=os.path.basename(full_filename),
            block_number=self.entry.id, msg_type='upload_file')
        cloud.send_message_and_wait(data)

    def verify_cloud_backup(self):
        """Check that cloud storage has this file with matching MD5 sum"""
        full_filename = self.get_path()

        import hashlib
        def md5(fname):
            hash_md5 = hashlib.md5()
            with open(fname, "rb") as f:
                for chunk in iter(lambda: f.read(4096), b""):
                    hash_md5.update(chunk)
            return hash_md5.hexdigest()

        data = dict(full_filename=full_filename, filename=os.path.basename(full_filename),
            md5_hash=md5(full_filename), block_number=self.entry.id, msg_type='check_status')
        cloud.send_message_and_wait(data)


class TaskEntryCollection(models.Model):
    """ Collection of TaskEntry records grouped together, e.g. for analysis """
    name = models.CharField(max_length=200, default='')
    entries = models.ManyToManyField(TaskEntry)

    @property
    def safe_name(self):
        return self.name.replace(' ', '_')

    def add_entry(self, te):
        if not isinstance(te, TaskEntry):
            te = TaskEntry(id=te)

        if te not in self.entries.all():
            self.entries.add(te)

    def remove_entry(self, te):
        if not isinstance(te, TaskEntry):
            te = TaskEntry(id=te)

        if te in self.entries.all():
            self.entries.remove(te)


class KeyValueStore(models.Model):
    key = models.TextField()
    value = models.TextField()

    def __str__(self):
        return "KV[%s => %s]" % (self.key, self.value)

    def __repr__(self):
        return self.__str__()

    @classmethod
    def get(cls, key, default=None, dbname=None):
        try:
            if dbname is not None:
                objs = cls.objects.using(dbname).filter(key=key)
            else:
                objs = cls.objects.filter(key=key)
            if len(objs) == 0:
                return default
            if len(objs) == 1:
                return objs[0].value
            if len(objs) > 1:
                raise ValueError("Duplicate keys: %s" % key)
        except:
            print("KeyValueStore error")
            print("key = %s, default=%s, dbname=%s" % (key, default, dbname))
            traceback.print_exc()
            return default

    @classmethod
    def set(cls, key, value, dbname=None):
        if dbname is not None:
            matching_recs = cls.objects.using(dbname).filter(key=key)
        else:
            matching_recs = cls.objects.filter(key=key)
        if len(matching_recs) == 0:
            obj = cls(key=key, value=value)
            obj.save()
        elif len(matching_recs) == 1:
            obj = matching_recs[0]
            obj.value = value
            obj.save()
        else:
            raise ValueError("Duplicate keys: %s" % key)




class Singleton(object):
    '''Python implementation of the Singleton design pattern'''
    __instance = None

    @classmethod
    def get_instance(cls):
        instance_attr_name = '_%s__instance' % cls.__name__
        if getattr(cls, instance_attr_name) is None:
            cls()
        return getattr(cls, instance_attr_name)

    def __init__(self, *args, **kwargs):
        instance_attr_name = '_%s__instance' % self.__class__.__name__
        if getattr(self.__class__, instance_attr_name) is None:
            setattr(self.__class__, instance_attr_name, self)
        else:
            raise Exception("can't reinstantiate %s because it is a singleton!" % self.__class__.__name__)




'''
BMI tasks in the new structure, i.e. inheriting from manualcontrolmultitasks
'''
import numpy as np
import pickle

from riglib.experiment import traits

from riglib.bmi import goal_calculators, ppfdecoder, feedback_controllers
from riglib.bmi.bmi import BMILoop
from riglib.bmi.assist import Assister, FeedbackControllerAssist
from riglib.bmi.state_space_models import StateSpaceEndptVel2D, StateSpaceNLinkPlanarChain
from riglib.experiment.experiment import control_decorator

from riglib.stereo_opengl.window import WindowDispl2D
from .target_capture_task import ScreenReachAngle, ScreenTargetCapture
from features.bmi_task_features import LinearlyDecreasingAssist
from .target_graphics import target_colors

np.set_printoptions(suppress=False)

###################
####### Assisters
##################
class OFCEndpointAssister(FeedbackControllerAssist):
    '''
    Assister for cursor PPF control which uses linear feedback (infinite horizon LQR) to drive the cursor toward the target state
    '''
    def __init__(self, decoding_rate=180):
        '''
        Constructor for OFCEndpointAssister

        Parameters
        ----------
        decoding_rate : int
            Rate that the decoder should operate, in Hz. Should be a multiple or divisor of 60 Hz

        Returns
        -------
        OFCEndpointAssister instance
        '''
        F_dict = pickle.load(open('/storage/assist_params/assist_20levels_ppf.pkl'))
        B = np.mat(np.vstack([np.zeros([3,3]), np.eye(3)*1000*1./decoding_rate, np.zeros(3)]))
        fb_ctrl = feedback_controllers.MultiModalLFC(A=B, B=B, F_dict=F_dict)
        super(OFCEndpointAssister, self).__init__(fb_ctrl, style='additive_cov')
        self.n_assist_levels = len(F_dict)

    def get_F(self, assist_level):
        '''
        Look up the feedback gain matrix based on the assist_level

        Parameters
        ----------
        assist_level : float
            Float between 0 and 1 to indicate the level of the assist (1 being the highest)

        Returns
        -------
        np.mat
        '''
        assist_level_idx = min(int(assist_level * self.n_assist_levels), self.n_assist_levels-1)
        F = np.mat(self.fb_ctrl.F_dict[assist_level_idx])
        return F

class SimpleEndpointAssister(Assister):
    '''
    Constant velocity toward the target if the cursor is outside the target. If the
    cursor is inside the target, the speed becomes the distance to the center of the
    target divided by 2.
    '''
    def __init__(self, *args, **kwargs):
        '''    Docstring    '''
        self.decoder_binlen = kwargs.pop('decoder_binlen', 0.1)
        self.assist_speed = kwargs.pop('assist_speed', 5.)
        self.target_radius = kwargs.pop('target_radius', 2.)
        self.assist_noise = kwargs.pop('assist_noise', 0.)

    def calc_assisted_BMI_state(self, current_state, target_state, assist_level, mode=None, **kwargs):
        '''    Docstring    '''
        Bu = None
        assist_weight = 0.

        if assist_level > 0:
            cursor_pos = np.array(current_state[0:3,0]).ravel()
            target_pos = np.array(target_state[0:3,0]).ravel()
            decoder_binlen = self.decoder_binlen
            speed = self.assist_speed * decoder_binlen
            target_radius = self.target_radius
            Bu = self.endpoint_assist_simple(cursor_pos, target_pos, decoder_binlen, speed, target_radius, self.assist_noise)
            assist_weight = assist_level

        # return Bu, assist_weight
        return dict(x_assist=Bu, assist_level=assist_weight)

    @staticmethod
    def endpoint_assist_simple(cursor_pos, target_pos, decoder_binlen=0.1, speed=0.5, target_radius=2., assist_noise=0.):
        '''
        Estimate the next state using a constant velocity estimate moving toward the specified target

        Parameters
        ----------
        cursor_pos: np.ndarray of shape (3,)
            Current position of the cursor
        target_pos: np.ndarray of shape (3,)
            Specified target position
        decoder_binlen: float
            Time between iterations of the decoder
        speed: float
            Speed of the machine-assisted cursor
        target_radius: float
            Radius of the target. When the cursor is inside the target, the machine assisted cursor speed decreases.
        assist_noise: float
            Noise added to the assist speed to vary the timing of the trajectories

        Returns
        -------
        x_assist : np.ndarray of shape (7, 1)
            Control vector to add onto the state vector to assist control.
        '''
        diff_vec = target_pos - cursor_pos
        dist_to_target = np.linalg.norm(diff_vec)
        dir_to_target = diff_vec / (np.spacing(1) + dist_to_target)

        if dist_to_target > target_radius:
            assist_cursor_pos = cursor_pos + np.random.uniform(1-assist_noise,1+assist_noise)*speed*dir_to_target
        else:
            assist_cursor_pos = cursor_pos + np.random.uniform(1-assist_noise,1+assist_noise)*speed*diff_vec/2

        assist_cursor_vel = (assist_cursor_pos-cursor_pos)/decoder_binlen
        x_assist = np.hstack([assist_cursor_pos, assist_cursor_vel, 1])
        x_assist = np.mat(x_assist.reshape(-1,1))
        return x_assist

class SimplePosAssister(SimpleEndpointAssister):
    
    @staticmethod 
    def endpoint_assist_simple(cursor_pos, target_pos, decoder_binlen=0.1, speed=0.5, target_radius=2., assist_level=0.):
        '''
        Estimate the next state using a constant velocity estimate moving toward the specified target

        Parameters
        ----------
        see SimpleEndtpointAssister for docs

        Returns
        -------
        x_assist : np.ndarray of shape (7, 1)
            Control vector to add onto the state vector to assist control.
        '''
        diff_vec = target_pos - cursor_pos 
        dist_to_target = np.linalg.norm(diff_vec)
        dir_to_target = diff_vec / (np.spacing(1) + dist_to_target)
        
        if dist_to_target > target_radius:
            assist_cursor_pos = cursor_pos + speed*dir_to_target
        else:
            assist_cursor_pos = cursor_pos + speed*diff_vec/2

        return assist_cursor_pos.ravel()

class SimpleEndpointAssisterLFC(feedback_controllers.MultiModalLFC):
    '''
    Docstring
    '''
    def __init__(self, *args, **kwargs):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        dt = 0.1
        A = np.mat([[1., 0, 0, dt, 0, 0, 0],
                    [0., 1, 0, 0,  dt, 0, 0],
                    [0., 0, 1, 0, 0, dt, 0],
                    [0., 0, 0, 0, 0,  0, 0],
                    [0., 0, 0, 0, 0,  0, 0],
                    [0., 0, 0, 0, 0,  0, 0],
                    [0., 0, 0, 0, 0,  0, 1]])

        I = np.mat(np.eye(3))
        B = np.vstack([0*I, I, np.zeros([1,3])])
        F_target = np.hstack([I, 0*I, np.zeros([3,1])])
        F_hold = np.hstack([0*I, 0*I, np.zeros([3,1])])
        F_dict = dict(hold=F_hold, target=F_target)
        super(SimpleEndpointAssisterLFC, self).__init__(B=B, F_dict=F_dict)

#################
##### Tasks #####
#################
class BMIControlMultiMixin(BMILoop, LinearlyDecreasingAssist):
    '''
    Target capture task with cursor position controlled by BMI output.
    Cursor movement can be assisted toward target by setting assist_level > 0.
    '''
    reset = traits.Int(0, desc='reset the decoder state to the starting configuration. 1 for always, 2 for only on timeout')
    assist_speed = traits.Float(2., desc="speed of assister in cm/s")
    assist_noise = traits.Float(0., desc="noise added to cursor speed in cm/s")
    cursor_color = traits.OptionsList("orange", *target_colors, desc='Color of cursor endpoint', bmi3d_input_options=list(target_colors.keys()))
    save_zscore = traits.Bool(False, desc="save a decoder zscored from this task")

    static_states = ['reward'] # states in which the decoder is not run

    ordered_traits = ['session_length', 'assist_level', 'assist_level_time', 'reward_time','timeout_time','timeout_penalty_time']
    exclude_parent_traits = ['marker_count', 'marker_num', 'goal_cache_block']
    hidden_traits = ['arm_hide_rate', 'arm_visible', 'hold_penalty_time', 'rand_start', 'reset', 'target_radius', 'window_size']

    is_bmi_seed = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    def create_assister(self):
        # Create the appropriate type of assister object
        start_level, end_level = self.assist_level
        kwargs = dict(decoder_binlen=self.decoder.binlen, target_radius=self.target_radius)
        kwargs['assist_speed'] = self.assist_speed
        kwargs['assist_noise'] = self.assist_noise

        if isinstance(self.decoder.ssm, StateSpaceEndptVel2D) and isinstance(self.decoder, ppfdecoder.PPFDecoder):
            self.assister = OFCEndpointAssister()
        elif isinstance(self.decoder.ssm, StateSpaceEndptVel2D):
            self.assister = SimpleEndpointAssister(**kwargs)
        ## elif (self.decoder.ssm == namelist.tentacle_2D_state_space) or (self.decoder.ssm == namelist.joint_2D_state_space):
        ##     # kin_chain = self.plant.kin_chain
        ##     # A, B, W = self.decoder.ssm.get_ssm_matrices(update_rate=self.decoder.binlen)
        ##     # Q = np.mat(np.diag(np.hstack([kin_chain.link_lengths, np.zeros_like(kin_chain.link_lengths), 0])))
        ##     # R = 10000*np.mat(np.eye(B.shape[1]))

        ##     # fb_ctrl = LQRController(A, B, Q, R)
        ##     # self.assister = FeedbackControllerAssist(fb_ctrl, style='additive')
        ##     self.assister = TentacleAssist(ssm=self.decoder.ssm, kin_chain=self.plant.kin_chain, update_rate=self.decoder.binlen)
        else:
            raise NotImplementedError("Cannot assist for this type of statespace: %r" % self.decoder.ssm)

        print('Assister: ', self.assister)

    def create_goal_calculator(self):
        if isinstance(self.decoder.ssm, StateSpaceEndptVel2D):
            self.goal_calculator = goal_calculators.ZeroVelocityGoal(self.decoder.ssm)
        elif isinstance(self.decoder.ssm, StateSpaceNLinkPlanarChain) and self.decoder.ssm.n_links == 2:
            self.goal_calculator = goal_calculators.PlanarMultiLinkJointGoal(self.decoder.ssm, self.plant.base_loc, self.plant.kin_chain, multiproc=False, init_resp=None)
        elif isinstance(self.decoder.ssm, StateSpaceNLinkPlanarChain) and self.decoder.ssm.n_links == 4:
            shoulder_anchor = self.plant.base_loc
            chain = self.plant.kin_chain
            q_start = self.plant.get_intrinsic_coordinates()
            x_init = np.hstack([q_start, np.zeros_like(q_start), 1])
            x_init = np.mat(x_init).reshape(-1, 1)

            cached = True

            if cached:
                goal_calc_class = goal_calculators.PlanarMultiLinkJointGoalCached
                multiproc = False
            else:
                goal_calc_class = goal_calculators.PlanarMultiLinkJointGoal
                multiproc = True

            self.goal_calculator = goal_calc_class(namelist.tentacle_2D_state_space, shoulder_anchor,
                                                   chain, multiproc=multiproc, init_resp=x_init)
        else:
            raise ValueError("Unrecognized decoder state space!")

    def get_target_BMI_state(self, *args):
        '''
        Run the goal calculator to determine the target state of the task
        '''
        if isinstance(self.goal_calculator, goal_calculators.PlanarMultiLinkJointGoalCached):
            task_eps = np.inf
        else:
            task_eps = 0.5
        ik_eps = task_eps/10
        data, solution_updated = self.goal_calculator(self.target_location, verbose=False, n_particles=500, eps=ik_eps, n_iter=10, q_start=self.plant.get_intrinsic_coordinates())
        target_state, error = data

        if isinstance(self.goal_calculator, goal_calculators.PlanarMultiLinkJointGoal) and error > task_eps and solution_updated:
            self.goal_calculator.reset()

        return np.array(target_state).reshape(-1,1)

    def _end_targ_transition(self):
        super()._end_targ_transition()
        if self.reset == 1 and ((self.target_index == self.chain_length - 1) or (self.target_index == -1)):

            # Reset on any target transition away from the last target
            self.decoder.filt.state.mean = self.init_decoder_mean.copy()
            self.hdf.sendMsg("reset")

    def _end_timeout_penalty(self):
        super()._end_timeout_penalty()
        if self.reset == 2:

            # Reset on timeout
            self.decoder.filt.state.mean = self.init_decoder_mean.copy()
            self.hdf.sendMsg("reset")

    def _start_None(self):
        super()._start_None()

        # Optionally save a new decoder zscored from this task
        if (not self.save_zscore) or (self.saveid is None):
            return

        if not (np.all(self.decoder.mFR == 0) and np.all(self.decoder.sdFR) == 1):
            filename = self.decoder.save()

            from db.tracker import dbq
            suffix = f"saved_from_{self.saveid}"
            dbq.save_bmi(suffix, self.saveid, filename)
            return

        # This is linear mapping specific - needs updating
        self.decoder.filt.fix_norm_attr()
        self.decoder.filt._update_scale_attr()
        mFR = self.decoder.filt.attr['offset'].copy()
        sdFR = self.decoder.filt.attr['scale'].copy()
        n_units = self.decoder.filt.n_units
        self.decoder.filt.update_norm_attr(offset=np.zeros(n_units), scale=np.ones(n_units))

        # The rest should work with any decoder
        self.decoder.init_zscore(mFR, sdFR)
        filename = self.decoder.save()

        from db.tracker import dbq
        suffix = f"zscored_online_in_{self.saveid}"
        dbq.save_bmi(suffix, self.saveid, filename)


    @classmethod
    def get_desc(cls, params, log_summary):
        duration = round(log_summary['runtime'] / 60, 1)
        return "{}/{} succesful trials in {} min".format(
            log_summary['n_success_trials'], log_summary['n_trials'], duration)

    @control_decorator
    def reset_cursor(self):
        self.decoder.filt.state.mean = self.init_decoder_mean.copy()
        self.hdf.sendMsg("reset")
        
    # @control_decorator
    # def update_zscore(self):
    #     # This is a temporary solution - LRS Jan 2024
    #     # Only works if the decoder currently has mFR == 0 and sdFR == 1
    #     # Assumes you're using a lindecoder with a reasonably large buffer (>2 minutes)
    #     self.decoder.filt.fix_norm_attr() # Should be already!
    #     self.decoder.filt._update_scale_attr()
    #     print(f"updated zscore")
    #     self.hdf.sendMsg(f"updated zscore")
    #     mFR = self.decoder.filt.attr['offset'].copy()
    #     sdFR = self.decoder.filt.attr['scale'].copy()
    #     self.decoder.init_zscore(mFR, sdFR)
    #     self.hdf.sendAttr("task", "session_mFR", mFR)
    #     self.hdf.sendAttr("task", "session_sdFR", sdFR)
    #     n_units = self.decoder.filt.n_units
    #     self.decoder.filt.update_norm_attr(offset=np.zeros(n_units), scale=np.ones(n_units))

    @control_decorator
    def toggle_clda(self):
        self.learn_flag = not self.learn_flag
        self.hdf.sendMsg(f"clda = {self.learn_flag}")
        print(f"clda = {self.learn_flag}")

class BMIControlMulti2DWindow(BMIControlMultiMixin, WindowDispl2D, ScreenTargetCapture):
    fps = 20.
    def __init__(self,*args, **kwargs):
        super(BMIControlMulti2DWindow, self).__init__(*args, **kwargs)

    def create_assister(self):
        kwargs = dict(decoder_binlen=self.decoder.binlen, target_radius=self.target_radius)
        if hasattr(self, 'assist_speed'):
            kwargs['assist_speed'] = self.assist_speed
        self.assister = SimpleEndpointAssister(**kwargs)

    def create_goal_calculator(self):
        self.goal_calculator = goal_calculators.ZeroVelocityGoal(self.decoder.ssm)

    def _start_wait(self):
        self.wait_time = 0.
        super(BMIControlMulti2DWindow, self)._start_wait()

    def _test_start_trial(self, ts):
        return ts > self.wait_time and not self.pause

class BMIControlMulti(BMIControlMultiMixin, ScreenTargetCapture):
    '''
    Slightly refactored original bmi control task
    '''
    pass

class BMIControlMultiDirectionConstraint(BMIControlMultiMixin, ScreenReachAngle):
    '''
    Adds an additional constraint that the direction of travel must be within a certain angle
    '''
    pass


from riglib.bmi.extractor import LFPMTMPowerExtractor
import numpy as np

import unittest

class TestMTM(unittest.TestCase):

    extractor = LFPMTMPowerExtractor(None, [0,1], bands=[(80,100),(100,150)], NW=3)
    print(extractor.npk)
    cont_samples = np.random.normal((2,200))
    feats = extractor.extract_features(cont_samples)
    
if __name__ == '__main__':
    unittest.main()





import numpy as np


class RingBuffer(object):

    def __init__(self, item_len, capacity):
        self.item_len = item_len
        self.capacity = capacity
        self.buffer = np.zeros((item_len, capacity))
        self.n_items = 0
        self.idx = 0

    def add(self, item):
        self.buffer[:, self.idx] = item
        self.idx = (self.idx + 1) % self.capacity
        self.n_items = min(self.n_items + 1, self.capacity)

    def add_multiple_values(self, item):
        l = item.shape[1]    
        
        if (self.idx + l) <= self.capacity:
            self.buffer[:, self.idx:self.idx + l] = item
            self.idx = (self.idx + l) % self.capacity
        else:
            self.new_idx =  (self.idx + l) % self.capacity
            self.buffer[:, self.idx:] = item[:,0:l-self.new_idx]
            self.buffer[:, :self.new_idx] = item[:,l-self.new_idx:]
            self.idx = self.new_idx
        self.n_items = min(self.n_items + l, self.capacity)
        
    def get(self, n_items):
        if n_items > self.n_items:
            return None
        else:
            if n_items <= self.idx:
                return self.buffer[:, self.idx-n_items:self.idx]
            else:
                return np.hstack([self.buffer[:, -(n_items-self.idx):], self.buffer[:, :self.idx]])

    def get_all(self):
        if self.n_items < self.capacity:
            return self.buffer[:, :self.idx]
        else:
            return np.hstack([self.buffer[:, self.idx:], self.buffer[:, :self.idx]])

    def is_full(self):
        return self.n_items == self.capacity

    def num_items(self):
        return self.n_items



__doc__ = """
The Robotics Toolbox for Python
Based on the Matlab version
Peter Corke 2007
"""

######################
#   Import Section   #
######################

# Import Link Constructor section
from .Link import *

# Import Robot Constructor section
from .SerialLink import *



#!/usr/bin/python
"""
State-space models for different types of BMIs (i.e. representations of various 
BMI "plants"), and methods to manipulate the parameterizations of such models 
e.g. time resampling of a linear discrete-time representation of a continuous-time model). 
"""
import numpy as np

############
## Constants
############
pi = np.pi


class State(object):
    '''
    A 1D component of a state-space, e.g., vertical velocity
    '''
    def __init__(self, name, stochastic=False, drives_obs=False, min_val=np.nan, max_val=np.nan, order=-1, aux=False):
        '''
        Constructor for State

        Parameters
        ----------
        name : string
            Name of the state
        stochastic : bool, optional
            Specify whether the state is stochastic (estimated from observation) 
            or deterministic (updated entirely by model). Default is 'deterministic'
        drives_obs : bool, optional
            Specify whether the state will be reflected in the observations if it is
            used as a 'hidden' state. By default, the state and any observations will not be directly related
        min_val : float, optional
            Hard (nonlinear) constraint on the minimum value of the state. By default (np.nan), no constraint is applied.
        max_val : float, optional
            Hard (nonlinear) constraint on the maximum value of the state. By default (np.nan), no constraint is applied.
        order : int
            Specification of the 'order' that this state would contribute to a differential equation, 
            e.g., integrated position is order -1, position states are order 0, velocity states are order 1, acceleration is order 2, etc.
            Constant states should be order NaN

        Returns
        -------
        State instance
        '''
        assert not name == 'q', "'q' is a reserved keyword (symbol for generalized robot coordinates) and cannot be used as a state name"
        self.name = name
        self.stochastic = stochastic 
        self.drives_obs = drives_obs
        self.min_val = min_val
        self.max_val = max_val
        self.order = order
        self.aux = aux
        self._eq_comp_excl = []

    def __repr__(self):
        return str(self.name) 

    def __eq__(self, other):
        '''
        State instances are equal if all their attributes (name, stochastic, etc.) are equal
        '''
        if not isinstance(other, State):
            return False
        else:
            for x in self.__dict__:
                if x in other.__dict__:
                    if not (x == '_eq_comp_excl') and not (x in self._eq_comp_excl) and not (x in other._eq_comp_excl):
                        if not (self.__dict__[x] == other.__dict__[x] or (np.isnan(self.__dict__[x]) and np.isnan(other.__dict__[x]))):
                            print(self, other, x)
                            import pdb; pdb.set_trace()
                            return False
            return True

    def __setstate__(self, data):
        self.__dict__ = data
        if '_eq_comp_excl' not in self.__dict__:
            self._eq_comp_excl = []


class StateSpace(object):
    '''
    A collection of multiple 'State' instances forms a StateSpace
    '''
    def __init__(self, *states, **kwargs):
        '''
        Constructor for StateSpace

        Parameters
        ----------
        states : packed tuple
            State instances specified in comma-separated arguments

        Returns
        -------
        StateSpace instance
        '''
        if 'statelist' in kwargs:
            self.states = kwargs['statelist']
        else:
            self.states = list(states)

    def __len__(self):
        return len(self.states)
        
    def __repr__(self):
        return 'State space: ' + str(self.state_names)

    @property
    def is_stochastic(self):
        '''
        An array of booleans specifying each state as stochastic
        '''
        return np.array([x.stochastic for x in self.states])

    @property
    def drives_obs(self):
        '''
        An array of booleans specifying each state as observation-driving
        '''
        return np.array([x.drives_obs for x in self.states])

    @property 
    def is_aux_state(self):
        return np.array([x.aux for x in self.states])

    @property
    def state_names(self):
        '''
        A list of string names for each state
        '''
        return [x.name for x in self.states]

    @property
    def bounding_box(self):
        '''
        A tuple of min values and max values for each state
        '''
        min_bounds = np.array([x for x in [x.min_val for x in self.states] if x is not np.nan])
        max_bounds = np.array([x for x in [x.max_val for x in self.states] if x is not np.nan])
        return (min_bounds, max_bounds)

    @property
    def states_to_bound(self):
        '''
        A list of the names of all the states which have limits on the values they can take.
        '''
        return [x.name for x in [x for x in self.states if x.min_val is not np.nan]]

    @property
    def n_states(self):
        '''
        Number of states in the space
        '''
        return len(self.states)

    @property
    def train_inds(self):
        '''
        An array of 
        '''
        return [k for k in range(self.n_states) if self.states[k].stochastic]

    @property
    def drives_obs_inds(self):
        '''
        A list of the indices of the states which are related to observations when 
        used as a hidden state-space. Used when seeding Decoders
        '''
        return [k for k in range(self.n_states) if self.states[k].drives_obs]

    @property 
    def state_order(self):
        '''
        An array listing the 'order' of each state (see State.__init__ for description of 'order')
        '''
        return np.array([x.order for x in self.states])

    def get_ssm_matrices(self, *args, **kwargs):
        '''
        Returns the parameters of the composite state-space models for use in Decoders. 
        Must be overridden in child classes as there is no way to specify this generically.
        '''
        raise NotImplementedError

    def __eq__(self, other):
        '''
        State spaces are equal if the all the states are equal and all the states are listed in the same order
        '''
        if not isinstance(other, StateSpace):
            return False
        else:
            return self.states == other.states


class LinearVelocityStateSpace(StateSpace):
    def __init__(self, states, vel_decay=0.8, w=7, Delta=0.1):
        self.states = states 
        self.vel_decay = vel_decay
        self.w = w
        self.Delta = Delta

        # check that there are an equal number of pos and vel states
        assert len(np.nonzero(self.state_order == 0)[0]) == len(np.nonzero(self.state_order == 1)[0])

    def __setstate__(self, state):
        self.__dict__ = state
        if not hasattr(self, 'Delta'):
            self.Delta = 0.1

        if not hasattr(self, 'vel_decay'):
            self.vel_decay = 0.8

        if not hasattr(self, 'w'):
            self.w = 7

    def get_ssm_matrices(self, update_rate=0.1):
        '''
        For the linear stochastic state-space model 
            x_{t+1} = Ax_{t} + Bu_t + w_t;   w_t ~ N(0, W),
        this function specifies the matrices A, B and W

        A = [I_N    \Delta I_N   0
             0_N    a*I_N        0
             0      0            1]

        W = [0_N    0_N   0
             0_N    w*I_N        0
             0      0            0]

        B = [0_N
            1000\Delta I_N
            0]

        Parameters
        ----------
        update_rate : float, optional
            Time between iterations of the discrete-time model. Default is 0.1 sec.

        Returns
        -------
        tuple of 3 np.mat matrices
            A, B and W as specified in the mathematical model above
        '''
        if not (update_rate is None):
            a_resamp, w_resamp = resample_scalar_ssm(self.vel_decay, self.w, Delta_old=self.Delta, Delta_new=update_rate)
            Delta = update_rate
        else:
            a_resamp = self.vel_decay
            w_resamp = self.w
            Delta = self.Delta

        ndim = len(np.nonzero(self.state_order == 1)[0])
        A = _gen_A(1, Delta, 0, a_resamp, 1, ndim=ndim)
        W = _gen_A(0, 0, 0, w_resamp, 0, ndim=ndim)        

        # Control input matrix for SSM for control inputs
        I = np.mat(np.eye(ndim))
        B = np.vstack([0*I, Delta*1000 * I, np.zeros([1, ndim])])

        # account for offset state
        has_offset = self.states[-1] == offset_state
        # has_offset = np.isnan(self.states[-1].order)
        if not has_offset:
            A = A[:-1, :-1]
            W = W[:-1, :-1]
            B = B[:-1, :]

        return A, B, W

    def __eq__(self, other):
        states_equal = super(LinearVelocityStateSpace, self).__eq__(other)
        A1, B1, W1 = self.get_ssm_matrices()
        A2, B2, W2 = other.get_ssm_matrices()
        # import pdb; pdb.set_trace()
        return states_equal and np.array_equal(A1, A2) and np.array_equal(B1, B2) and np.array_equal(W1, W2)


#######################################################################
##### Specific StateSpace types for particular experiments/plants #####
#######################################################################
# These class declarations may not actually be best placed in this class, 
# but moving them now would cause problems with unpickling older decoder 
# objects, which are saved with these state space model types. So put new 
# state-space models elsewhere!
offset_state = State('offset', stochastic=False, drives_obs=True, order=np.nan)
offset_state._eq_comp_excl.append('order')

class StateSpaceNLinkPlanarChain(LinearVelocityStateSpace):
    '''
    State-space model for an N-link kinematic chain
    '''
    def __init__(self, n_links=2, **kwargs):
        self.n_links = n_links
        pos_states = []
        vel_states = []

        for k in range(n_links):
            pos_state_k = State('theta_%d' % k, stochastic=False, drives_obs=False, min_val=-pi, max_val=0, order=0)
            vel_state_k = State('omega_%d' % k, stochastic=True, drives_obs=True, order=1)
            pos_states.append(pos_state_k)
            vel_states.append(vel_state_k)

        states = pos_states + vel_states + [offset_state]
        super(StateSpaceNLinkPlanarChain, self).__init__(states, **kwargs)

    def __setstate__(self, state):
        self.__dict__ = state
        if not hasattr(self, 'Delta'):
            self.Delta = 0.1

        if not hasattr(self, 'vel_decay'):
            self.vel_decay = 0.8

        if not hasattr(self, 'w'):
            self.w = 0.01

class StateSpaceEndptVel2D(LinearVelocityStateSpace):
    '''
    StateSpace with 2D velocity in the X-Z plane
    '''
    def __init__(self, **kwargs):
        states = [
            State('hand_px', stochastic=False, drives_obs=False, min_val=-25., max_val=25., order=0),
            State('hand_py', stochastic=False, drives_obs=False, order=0),
            State('hand_pz', stochastic=False, drives_obs=False, min_val=-14., max_val=14., order=0),
            State('hand_vx', stochastic=True,  drives_obs=True, order=1),
            State('hand_vy', stochastic=False, drives_obs=False, order=1),
            State('hand_vz', stochastic=True,  drives_obs=True, order=1),
            offset_state]
        super(StateSpaceEndptVel2D, self).__init__(states, **kwargs)

    def __setstate__(self, state):
        self.__dict__ = state
        if not hasattr(self, 'Delta'):
            self.Delta = 0.1

        if not hasattr(self, 'vel_decay'):
            self.vel_decay = 0.8

        if not hasattr(self, 'w'):
            self.w = 7

class StateSpaceEndptVel3D(LinearVelocityStateSpace):
    def __init__(self, **kwargs):
        states = [
            State('hand_px', stochastic=False, drives_obs=False, min_val=-25., max_val=25., order=0),
            State('hand_py', stochastic=False, drives_obs=False, order=0),
            State('hand_pz', stochastic=False, drives_obs=False, min_val=-14., max_val=14., order=0),
            State('hand_vx', stochastic=True,  drives_obs=True, order=1),
            State('hand_vy', stochastic=True,  drives_obs=True, order=1),
            State('hand_vz', stochastic=True,  drives_obs=True, order=1),
            offset_state]
        super(StateSpaceEndptVel3D, self).__init__(states, **kwargs)  
    def __setstate__(self, state):
        self.__dict__ = state
        if not hasattr(self, 'Delta'):
            self.Delta = 0.1

        if not hasattr(self, 'vel_decay'):
            self.vel_decay = 0.8

        if not hasattr(self, 'w'):
            self.w = 7  

class StateSpaceEndptPos1D(StateSpace):
    ''' StateSpace for 1D pos control (e.g. RatBMI)'''
    def __init__(self, **kwargs):
        states = State('cursor_p', stochastic=False, drives_obs=True, min_val=-10e6, max_val=10e6, order=0)
            
        super(StateSpaceEndptPos1D, self).__init__(states, **kwargs)

    def __setstate__(self, state):
        self.__dict__ = state
        if not hasattr(self, 'Delta'):
            self.Delta = 0.1

        if not hasattr(self, 'vel_decay'):
            self.vel_decay = 0.8

        if not hasattr(self, 'w'):
            self.w = 7

############################
##### Helper functions #####
############################
def resample_ssm(A, W, Delta_old=0.1, Delta_new=0.005, include_offset=True):
    '''
    Change the effective sampling rate of a linear random-walk discrete-time state-space model
    That is, state-space models of the form 

    x_{t+1} = Ax_t + w_t, w_t ~ N(0, W)

    Parameters
    ----------
    A : np.mat of shape (K, K)
        State transition model
    W : np.mat of shape (K, K)
        Noise covariance estimate
    Delta_old : float, optional, default=0.1
        Old sampling rate
    Delta_new : float, optional, default=0.005
        New sampling rate
    include_offset : bool, optional, default=True
        Indicates whether the state-space matrices 

    Returns
    -------
    A_new, W_new
        New state-space model parameters at the new sampling rate.

    '''
    A = A.copy()
    W = W.copy()
    if include_offset:
        orig_nS = A.shape[0]
        A = A[:-1, :-1]
        W = W[:-1, :-1]

    loop_ratio = Delta_new/Delta_old
    N = 1./loop_ratio
    A_new = A**loop_ratio
    nS = A.shape[0]
    I = np.mat(np.eye(nS))
    W_new = W * ( (I - A_new**N) * (I - A_new).I - I).I
    if include_offset:
        A_expand = np.mat(np.zeros([orig_nS, orig_nS]))
        A_expand[:-1,:-1] = A_new
        A_expand[-1,-1] = 1
        W_expand = np.mat(np.zeros([orig_nS, orig_nS]))
        W_expand[:-1,:-1] = W_new
        return A_expand, W_expand
    else:
        return A_new, W_new

def resample_scalar_ssm(a, w, Delta_old=0.1, Delta_new=0.005):
    '''
    Similar to resample_ssm, but for a scalar (1-d) state-space model, 
    where the problem can be solved without complicated matrix roots

    Parameters
    ----------
    a : float
        State transition model
    w : float 
        Noise variance estimate
    Delta_old : float, optional, default=0.1
        Old sampling rate
    Delta_new : float, optional, default=0.005
        New sampling rate

    Returns
    -------
    a_new, w_new
        New state-space model parameters at the new sampling rate.

    '''
    loop_ratio = Delta_new/Delta_old
    a_delta_new = a**loop_ratio
    w_delta_new = w / ((1-a_delta_new**(2*(1./loop_ratio)))/(1- a_delta_new**2))

    mu = 1
    sigma = 0
    for k in range(int(1./loop_ratio)):
        mu = a_delta_new*mu
        sigma = a_delta_new * sigma * a_delta_new + w_delta_new
    return a_delta_new, w_delta_new

def _gen_A(t, s, m, n, off, ndim=3):
    """
    Utility function for generating block-diagonal matrices
    used by the KF
    
        [t*I, s*I, 0
         m*I, n*I, 0
         0,   0,   off]

    Parameters
    ----------
    t : float 
        See matrix equation above
    s : float 
        See matrix equation above
    m : float 
        See matrix equation above
    n : float 
        See matrix equation above
    off : float 
        See matrix equation above
    ndim : int, optional, default = 3 
        Number of states in each block, e.g. 3-states for (x,y,z) position

    Returns
    -------
    np.mat of shape (N, N); N = 2*ndim + 1
    """
    A = np.zeros([2*ndim+1, 2*ndim+1])
    A_lower_dim = np.array([[t, s], [m, n]])
    A[0:2*ndim, 0:2*ndim] = np.kron(A_lower_dim, np.eye(ndim))
    A[-1,-1] = off
    return np.mat(A)


if __name__ == '__main__':
    a_10hz = 0.8
    w_10hz = 0.0007

    Delta_old = 0.1
    Delta_new = 1./60
    a_60hz, w_60hz = resample_scalar_ssm(a_10hz, w_10hz, Delta_old=Delta_old, Delta_new=Delta_new)


import warnings
import requests as re
import urllib3
import json
import time

USERNAME = 'python'
PASSWORD = 'python'
WATCHTOWERURL = 'https://10.155.206.1:4343'
# CAMERA_SID_LIST = ['e3v822d','e3v821f']#,'e3v821b']
IFACE = ''
CONFIG = '480p15'
CODEC = 'H264'
ANNOTATION = 'None'
SEGTIME = '30m'
TIMEOUT = 25 # (s), timeout to check if 

class E3VisionInterface(object):
    """E3VisionInterface

    White-Matter e3Vision camera system interface class.

    Designed for the aoLab booth camera setup. Configure cameras in whitematter GUI before experiment.

    Example use:

    e3vi = E3VisionInterface()
    e3vi.update_camera_status()
    e3vi.start_rec()
    ...
    e3vi.stop_rec()

    """
    def __init__(self,session_name=None):
        self.username = USERNAME
        self.password = PASSWORD
        self.watchtowerurl = WATCHTOWERURL
        # self.camera_sid_list = CAMERA_SID_LIST
        self._create_session_subdir(session_name)
        self.iface = IFACE
        self.config = CONFIG
        self.codec = CODEC
        self.anno = ANNOTATION
        self.segtime = SEGTIME
        urllib3.disable_warnings(
            urllib3.exceptions.InsecureRequestWarning
        )
        self.apitoken = self.api_login()
        # self.configure_cameras()

    def _create_session_subdir(self,session_name):
        """_create_session_subdir

        Creates a subdirectory string for the current session. All video files from this object are saved to this subdirectory.

        Args:
            session_name (str): BMI3D session name.
        """
        if session_name is None:
            session_name = 'test'
        self.subdir = session_name

    def api_post(self,api_call_str,**kwargs): #TODO: This is completely bare. Put some exception handling on this vis a vis HTTP code (401, 404, etc)
        """api_post

        Wrapper method for all e3Vision Watchtower API POST re.

        Args:
            api_call_str (str): HTTP API request string. Defines request type. Examples: '/api/login', '/api/cameras/action'
            **kwargs: keyword and value pairs populating request data dict.

        Returns:
            r (request.Request): API call return in re.Request format.
        """
        api_call = self.watchtowerurl + api_call_str
        data = {}
        for k, v in kwargs.items():
            data_key = k + '[]' if isinstance(v,list) else k
            data[data_key] = v
        if hasattr(self,'apitoken'):
            data['apitoken'] = self.apitoken
        r = re.post(
            api_call,
            data = data,
            verify = False,
            timeout = 5,
        )
        r.raise_for_status()
        return r

    def api_get(self,api_call_str,**kwargs):
        """api_get
        
        Wrapper method for all e3Vision Watchtower API GET requests.

        Args:
            api_call_str (str): HTTP API request string.
            **kwargs: keyword and value pairs populating request param dict.

        Returns:
            r (request.Request): API call return.

        """
        api_call = self.watchtowerurl + api_call_str
        params = {}
        for k, v in kwargs.items():
            param_key = k + '[]' if isinstance(v,list) else k
            params[param_key] = v
        if hasattr(self,'apitoken'):
            params['apitoken'] = self.apitoken
        r = re.get(
            api_call,
            params=params,
            verify=False,
            timeout=5,
        )
        r.raise_for_status()
        return r

    def api_login(self):
        """get_api_token

        Opens an API session with Watchtower. Returns the API auth token for the current session.

        Returns:
            apitoken (str): auth token for the current session. 
        """
        r = self.api_post(
            '/api/login',
            username=self.username,
            password=self.password,
        )
        j = json.loads(r.text)
        return j['apitoken']

    def update_camera_status(self):
        """update_camera_status

        Gets current camera list and updates self.camera_list
        """
        self.api_get(
            '/api/cameras/scan'
        )
        r_cameras = self.api_get(
            '/api/cameras'
        )
        self.camera_list = json.loads(r_cameras.text)
        # for cam in self.camera_list:
        #     print(f"Camera available: {cam['Id']}")

    def configure_cameras(self):
        """configure_cameras

        Binds, syncs and connects to all cameras in interface serial ID list using the desired configuration.
        """
        for cam in self.camera_list:
            id = cam['Id']
            self._bind_camera(id)
            self._update_sync(id)
            self._connect_camera(id)

    def start_rec(self, force=False):
        """start_rec

        Begin recording a video file to the session subdirectory. Records from all connected cameras simulataneously to separate files.
        File names have the following form: <global_dir>/<session_subdir>/[cameraname]-[starttime]-[endtime].[avi | mp4]
        """
        connected_camera_list = [cam for cam in self.camera_list if cam['Syncstate'] == 1 and cam['Alivestate'] > 0]
        rec_camera_list = [cam['Id'] for cam in connected_camera_list]
        rec_state = [cam['Recordstate'] for cam in connected_camera_list]
        if any(rec_state):
            if force:
                rec_warning_msg = 'Cameras already recording video. Stopping current recording before starting new recording.'
                warnings.warn(rec_warning_msg, RuntimeWarning)
            else:
                rec_error_msg = 'Cameras already recording video. Aborting recording session initiation. Enable overwrite behavior with force=True.'
                raise RuntimeError(rec_error_msg)
        self.api_post(
            '/api/cameras/action',
            IdGroup=rec_camera_list,
            Action='RECORDGROUP',
            AdditionalPath=self.subdir,
        )
        # check to see cameras are recording
        rec_check = True
        while rec_check:
            t_ping = time.time()
            self.update_camera_status()
            camera_running = [cam['Recordstate'] for cam in self.camera_list if cam['Id'] in rec_camera_list]
            rec_check = (not all(camera_running)) and (time.time() - t_ping < TIMEOUT)
        assert all(camera_running), 'Error starting video recordings.'
        print(f"Started e3v recordings on cameras {rec_camera_list}")

    def stop_rec(self):
        """stop_rec

        Stop all current video file recordings.
        """
        self.api_post(
            '/api/cameras/action',
            IdGroup=[cam['Id'] for cam in self.camera_list if cam['Recordstate']],
            Action='STOPRECORDGROUP',
        )
        print(f"Stopped e3v recordings")


    def _bind_camera(self,cid):
        """_bind_camera

        Binds the camera with serial ID cs. This establishes a secure authenticated session with the camera.

        Args:
            cid (str): Camera ID.
        """
        self.api_post(
            '/api/cameras/action',
            Id=cid,
            Action='BIND',
        )

    def _update_sync(self,cid):
        """_update_sync

        Updates camera synchronization for camera with serial ID cs.

        Args:
            cid (str): Camera ID.
        """
        self.api_post(
            '/api/cameras/action',
            Id=cid,
            Action='UPDATEMC',
        )
    
    def _connect_camera(self,cid):
        """_connect_camera

        Connects to camera with serial ID cs according to the specified configuration.

        Args:
            cid (str): Camera serial ID number.
        """
        self.api_post(
            '/api/cameras/action',
            Id=cid,
            Action='CONNECT',
            Iface='',
            Config=self.config,
            Codec=self.codec,
            Annotation=self.anno,
            Segtime=self.segtime,
        )



'''
Base tasks for generic point-to-point reaching
'''
import numpy as np
from riglib.stereo_opengl.primitives import Cable, Sphere, Cube, Torus
from riglib.stereo_opengl.primitives import Cylinder, Plane, Sphere, Cube
from riglib.stereo_opengl.models import FlatMesh, Group
from riglib.stereo_opengl.textures import Texture, TexModel
from riglib.stereo_opengl.render import stereo, Renderer
from riglib.stereo_opengl.utils import cloudy_tex

####### CONSTANTS
sec_per_min = 60.0
RED = (1,0,0,.5)
GREEN = (0,1,0,0.5)
BLUE = (0.,0.,1.,0.75)
GOLD = (1., 0.843, 0., 0.5)
YELLOW = (1,1,0,0.75)
mm_per_cm = 1./10

target_colors = {
    "red": (1,0,0,0.75),
    "yellow": (1,1,0,0.75),
    "green":(0., 1., 0., 0.75),
    "blue":(0.,0.,1.,0.75),
    "cyan":(0.,1.,1.,0.75),
    "magenta": (1,0,1,0.75),
    "pink": (1,0.5,1,0.75),
    "purple":(0.608,0.188,1,0.75),
    "dark_purple": (0.598,0.555,0.762,0.75),
    "teal":(0,0.502,0.502,0.75),
    "olive":(0.420,0.557,0.137,.75),
    "orange": (1,0.502,0.,0.75),
    "hotpink":(1,0.0,0.606,.75),
    "gold": (0.941,0.637,0.25,0.75),
    "elephant":(0.5,0.5,0.5,0.5),
    "white": (1, 1, 1, 0.75),
}

class CircularTarget(object): 
    def __init__(self, target_radius=2, target_color=(1, 0, 0, .5), starting_pos=np.zeros(3)):
        self.target_color = target_color
        self.target_radius = target_radius
        self.target_color = target_color
        self.position = starting_pos
        self.int_position = starting_pos
        self._pickle_init()

    def _pickle_init(self):
        self.sphere = Sphere(radius=self.target_radius, color=self.target_color)
        self.graphics_models = [self.sphere]
        self.sphere.translate(*self.position)

    def move_to_position(self, new_pos):
        self.int_position = new_pos
        self.drive_to_new_pos()

    def drive_to_new_pos(self):
        raise NotImplementedError

class VirtualCircularTarget(CircularTarget):
    def drive_to_new_pos(self):
        self.position = self.int_position
        self.sphere.translate(*self.position, reset=True)

    def hide(self):
        self.sphere.detach()

    def show(self):
        self.sphere.attach()

    def cue_trial_start(self):
        self.sphere.color = self.target_color
        self.show()

    def cue_fixation(self):
        self.sphere.color = BLUE

    def cue_trial_end_success(self):
        self.sphere.color = GREEN

    def cue_trial_end_failure(self):
        self.sphere.color = RED

    def pt_inside(self, pt):
        '''
        Test if a point is inside the target
        '''
        pos = self.sphere.xfm.move
        return (np.abs(pt[0] - pos[0]) < self.target_radius) and (np.abs(pt[2] - pos[2]) < self.target_radius)

    def reset(self):
        self.sphere.color = self.target_color

    def get_position(self):
        return self.sphere.xfm.move

class RectangularTarget(object):
    def __init__(self, target_width=4, target_height=4, target_color=(1, 0, 0, .5), starting_pos=np.zeros(3)):
        self.target_width = target_width
        self.target_height = target_height
        self.target_color = target_color
        self.position = starting_pos
        self.int_position = starting_pos
        self._pickle_init()

    def _pickle_init(self):
        self.cube = Cube(side_len=self.target_width,side_height=self.target_height,color=self.target_color)
        self.graphics_models = [self.cube]
        #self.center_offset = np.array([self.target_width, 0, self.target_width], dtype=np.float64) / 2
        self.center_offset = np.array([0, 0, self.target_width], dtype=np.float64) / 2
        corner_pos = self.position - self.center_offset
        self.cube.translate(*corner_pos)
    def move_to_position(self, new_pos):
        self.int_position = new_pos
        self.drive_to_new_pos()

    def drive_to_new_pos(self):
        raise NotImplementedError


class VirtualRectangularTarget(RectangularTarget):
    def drive_to_new_pos(self):
        self.position = self.int_position
        corner_pos = self.position - self.center_offset
        self.cube.translate(*corner_pos, reset=True)

    def hide(self):
        self.cube.detach()

    def show(self):
        self.cube.attach()

    def cue_trial_start(self):
        self.cube.color = RED
        self.show()

    def cue_trial_end_success(self):
        self.cube.color = GREEN

    def cue_trial_end_failure(self):
        self.cube.color = YELLOW
        self.hide()

    def idle(self):
        self.cube.color = RED
        self.hide()

    def pt_inside(self, pt):
        '''
        Test if a point is inside the target
        '''
        pos = self.cube.xfm.move + self.center_offset
        # TODO this currently assumes that the cube doesn't rotate
        # print (pt[0] - pos[0]), (pt[2] - pos[2])
        return (np.abs(pt[0] - pos[0]) < self.target_width/2) and (np.abs(pt[2] - pos[2]) < self.target_height/2)

    def reset(self):
        self.cube.color = self.target_color

    def get_position(self):
        return self.cube.xfm.move


class CableTarget(object): 
    def __init__(self, target_radius=1, target_color=(1, 0, 0, .5), starting_pos=np.zeros(3), trajectory = np.array([np.sin(x) for x in range(60)])):
        self.target_color = target_color
        self.target_radius = target_radius
        self.target_color = target_color
        self.position = starting_pos
        self.int_position = starting_pos
        self.trajectory = trajectory
        self._pickle_init()

    def _pickle_init(self):
        self.cable = Cable(radius=self.target_radius,trajectory = self.trajectory, color=self.target_color)
        self.graphics_models = [self.cable]
        self.cable.translate(*self.position)

    def move_to_position(self, new_pos):
        self.int_position = new_pos
        self.drive_to_new_pos()

    def drive_to_new_pos(self):
        raise NotImplementedError

        
class VirtualCableTarget(CableTarget):

    def update_shape(self):
        self.cable.trial_trajectory = self.trajectory
        self.cable.update()
    
    def drive_to_new_pos(self):
        self.position = self.int_position
        self.cable.translate(*self.position, reset=True)

    def hide(self):

        self.cable.detach()

    def show(self):      
        self.cable.attach()

    def cue_trial_start(self):
        self.cable.color = self.target_color
        self.show()

    def cue_trial_end_success(self):
        self.cable.color = GREEN

    def cue_trial_end_failure(self):
        self.cable.color = RED

    def pt_inside(self, pt):
        '''
        Test if a point is inside the target
        '''
        pos = self.cable.xfm.move
        return (np.abs(pt[0] - pos[0]) < self.target_radius) and (np.abs(pt[2] - pos[2]) < self.target_radius)

    def reset(self):
        self.cable.color = self.target_color

    def get_position(self):
        return self.cable.xfm.move

class VirtualTorusTarget(VirtualCircularTarget):

    def __init__(self, inner_radius=2, outer_radius=3, target_color=(1, 0, 0, .5), starting_pos=np.zeros(3)):
        self.target_color = target_color
        self.major_radius = np.mean([inner_radius, outer_radius])
        self.minor_radius = np.abs(outer_radius - inner_radius)/2
        self.target_color = target_color
        self.position = starting_pos
        self.int_position = starting_pos
        self._pickle_init()

    def _pickle_init(self):
        self.sphere = Torus(major_radius=self.major_radius, minor_radius=self.minor_radius, color=self.target_color)
        self.graphics_models = [self.sphere]
        self.sphere.translate(*self.position)
        self.sphere.rotate_x(90)

    def pt_inside(self, pt):
        '''
        Test if a point is inside the target
        '''
        pos = self.sphere.xfm.move
        # Not yet implmented. Hopefully not needed.
        return False



'''
Features for sending digital sync data
'''
from riglib.experiment import traits
from riglib.gpio import NIGPIO, ArduinoGPIO, DigitalWave, TeensyGPIO
import numpy as np
import tables
import time
import copy
import pygame
from built_in_tasks.target_graphics import VirtualRectangularTarget
from riglib.stereo_opengl.window import TRANSPARENT
from config.rig_defaults import hdf_sync_params, nidaq_sync_params, arduino_sync_params

def encode_event(dictionary, event_name, event_data):
    value = int(dictionary[event_name] + event_data)
    decoded = decode_event(dictionary, value)
    if decoded is None or decoded[0] != event_name or decoded[1] != event_data:
        raise Exception("Cannot encode " + event_name + " + " + str(event_data))
    return value

def decode_event(dictionary, value):
    ordered_list = sorted(dictionary.items(), key=lambda x: x[1])
    for i, event in enumerate(ordered_list[1:]):
        if value < event[1]:
            event_name = ordered_list[i][0]
            event_data = value - ordered_list[i][1]
            return event_name, event_data
    if value == ordered_list[-1][1]: # check last value
        return ordered_list[-1][0], 0
    return None

class HDFSync(traits.HasTraits):
    '''
    Sync events to the HDF file in a separate 'sync_events' dataset
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sync_params = hdf_sync_params
        self.sync_every_cycle = False

    def init(self, *args, **kwargs):

        # Create a record array for sync events
        if not hasattr(self, 'sinks'): # this attribute might be set in one of the other 'init' functions from other inherited classes
            from riglib import sink
            self.sinks = sink.SinkManager.get_instance()
        dtype = np.dtype([('time', 'u8'), ('event', 'S32'), ('data', 'u4'), ('code', 'u1')]) # 64-bit time (cycle number), string event, 32-bit data, 8-bit code
        self.sync_event_record = np.zeros((1,), dtype=dtype)
        self.sinks.register("sync_events", dtype)
        dtype = np.dtype([('time', 'u8'), ('timestamp', 'f8'), ('prev_tick', 'f8')])
        self.sinks.register("sync_clock", dtype)
        self.sync_clock_record = np.zeros((1,), dtype=dtype)
        super().init(*args, **kwargs)

        # Send a sync impulse to set t0
        time.sleep(self.sync_params['sync_pulse_width']*10)
        self.has_sync_event = False
        print("Sending t0 event")
        self.sync_event('TIME_ZERO', 0, immediate=True)
        self.t0 = time.perf_counter()
        time.sleep(self.sync_params['sync_pulse_width']*10)

    def sync_event(self, event_name, event_data=0, immediate=False):
        '''
        Send a sync event on the next cycle, unless 'immediate' flag is set
        '''
        super().sync_event(event_name, event_data=event_data, immediate=immediate)
        if self.has_sync_event:
            if self.sync_event_record['code'] == self.sync_params['event_sync_dict']['TRIAL_END'] and event_name == 'PAUSE_START':
                pass
            else:
                print("Warning: Cannot sync more than 1 event per cycle")
                print("Overwriting {} with {} event".format(self.sync_event_record['event'], event_name))

        # digital output
        code = encode_event(self.sync_params['event_sync_dict'], event_name, min(event_data, self.sync_params['event_sync_max_data']))
        self.sync_event_record['time'] = self.cycle_count
        self.sync_event_record['event'] = event_name
        self.sync_event_record['data'] = event_data
        self.sync_event_record['code'] = code
        if immediate:
            self.sinks.send("sync_events", self.sync_event_record)
            self.sync_code(int(self.sync_event_record['code']) << self.sync_params['event_sync_data_shift'])
            if hasattr(self, 'pulse'):
                self.pulse.join()
            self.has_sync_event = False
        else:
            self.has_sync_event = True

    def sync_code(self, code, delay=0.):
        '''
        Send a sync code through an digital io board
        '''
        pass

    def _cycle(self):
        '''
        Send a clock pulse on every cycle to the 'screen_sync_pin'. If there are any
        sync events, also send them in the same clock cycle.
        '''
        super()._cycle()
        code = 0
        if self.sync_every_cycle and 'screen_sync_pin' in self.sync_params:
            code = 1 << self.sync_params['screen_sync_pin']
        if self.has_sync_event:
            self.sinks.send("sync_events", self.sync_event_record)
            code |= int(self.sync_event_record['code']) << self.sync_params['event_sync_data_shift']
            self.has_sync_event = False
        if code > 0:
            self.sync_code(code)
        if self.sync_every_cycle:
            self.sync_clock_record['time'] = self.cycle_count
            self.sync_clock_record['timestamp'] = time.perf_counter() - self.t0
            self.sync_clock_record['prev_tick'] = self.clock.get_time()
            self.sinks.send("sync_clock", self.sync_clock_record)
            
    def cleanup_hdf(self):
        super().cleanup_hdf()
        if hasattr(self, "h5file"):
            h5file = tables.open_file(self.h5file.name, mode='a')
            for param in self.sync_params.keys():
                h5file.root.sync_events.attrs[param] = self.sync_params[param]
            h5file.close()

class NIDAQSync(HDFSync):
    '''
    Adds digital output to HDF sync to syncronize with an external recording device
    '''

    def __init__(self, *args, **kwargs):
        super(HDFSync, self).__init__(*args, **kwargs)
        self.sync_params = nidaq_sync_params
        self.sync_gpio = NIGPIO()
        self.sync_every_cycle = True
        print("NIDAQ sync active")

    def sync_code(self, code, delay=0.):
        '''
        Send a sync code through GPIO
        '''
        pulse = DigitalWave(self.sync_gpio, mask=self.sync_params['event_sync_mask'], data=code)
        pulse.set_edges([0, delay, self.sync_params['sync_pulse_width']], False)
        pulse.start()
        self.pulse = pulse

    def terminate(self):
        # Reset the sync pin after experiment ends so the next experiment isn't messed up
        self.sync_gpio.write_many(self.sync_params['event_sync_mask'], 0)

    def _cycle(self):
        # Wait for the previous sync to finish before starting this cycle! 
        # Important for preserving all the clock pulses, although it leads to slowdowns in framerate
        if hasattr(self, 'pulse'):
            self.pulse.join()
        super()._cycle()
            
class ArduinoSync(NIDAQSync):
    '''
    Use an arduino microcontroller to sync instead of a NI DIO card.
    '''

    sync_gpio_port = traits.String("/dev/teensydio", desc="Port used for digital sync")
    hidden_traits = ["sync_gpio_port"]
    
    def __init__(self, *args, **kwargs):
        super(HDFSync, self).__init__(*args, **kwargs)
        self.sync_params = arduino_sync_params
        self.sync_gpio = TeensyGPIO(self.sync_gpio_port)
        self.sync_every_cycle = True


class ScreenSync(traits.HasTraits):
    '''Adds a square in one corner that switches color with every flip.'''
    
    sync_position = {
        'TopLeft': (-1,1),
        'TopRight': (1,1),
        'BottomLeft': (-1,-1),
        'BottomRight': (1,-1)
    }
    sync_position_2D = {
        'TopLeft': (-1,-1),
        'TopRight': (1,-1),
        'BottomLeft': (-1,1),
        'BottomRight': (1,1)
    }
    sync_corner = traits.OptionsList(tuple(sync_position.keys()), desc="Position of sync square")
    sync_size = traits.Float(1, desc="Sync square size (cm)") 
    sync_color_off = traits.Tuple((0.,0.,0., 1.), desc="Sync off color (R,G,B,A)")
    sync_color_on = traits.Tuple((1.,1.,1., 1.), desc="Sync on color (R,G,B,A)")
    sync_state_duration = traits.Float(2., desc="How long to delay the start of the experiment (seconds)")
    sync_state_fps = traits.Float(30., desc="Frame rate during the sync state (lower helps to measure the screen latency)")

    hidden_traits = ['sync_color_off', 'sync_color_on', 'sync_state_duration', 'sync_state_fps']

    def __init__(self, *args, **kwargs):

        # Create a new "sync" state at the beginning of the experiment
        if isinstance(self.status, dict):
            self.status["sync"] = dict(start_experiment="wait", stoppable=False)
        else:
            from riglib.fsm.fsm import StateTransitions
            self.status.states["sync"] = StateTransitions(start_experiment="wait", stoppable=False)
        self.state = "sync"

        super().__init__(*args, **kwargs)
        self.sync_state = False
        if hasattr(self, 'is_pygame_display'):
            screen_center = np.divide(self.window_size,2)
            sync_size_pix = self.sync_size * self.window_size[0] / self.screen_cm[0]
            sync_center = [sync_size_pix/2, sync_size_pix/2]
            from_center = np.multiply(self.sync_position_2D[self.sync_corner], np.subtract(screen_center, sync_center))
            top_left = screen_center + from_center - sync_center
            self.sync_rect = pygame.Rect(top_left, np.multiply(sync_center,2))
        else:
            from_center = np.multiply(self.sync_position[self.sync_corner], np.subtract(self.screen_cm, self.sync_size))
            pos = np.array([from_center[0]/2, 1-self.screen_dist, from_center[1]/2])
            self.sync_square = VirtualRectangularTarget(target_width=self.sync_size, target_height=self.sync_size, target_color=self.sync_color_off, starting_pos=pos)
            # self.sync_square = VirtualCircularTarget(target_radius=self.sync_size, target_color=self.sync_color_off, starting_pos=pos)
            for model in self.sync_square.graphics_models:
                self.add_model(model)

    def screen_init(self):
        super().screen_init()
        if hasattr(self, 'is_pygame_display'):
            self.sync = pygame.Surface(self.window_size)
            self.sync.fill(TRANSPARENT)
            self.sync.set_colorkey(TRANSPARENT)

    def _draw_other(self):
        # For pygame display
        color = self.sync_color_on if self.sync_state else self.sync_color_off
        self.sync.fill(255*np.array(color), rect=self.sync_rect)
        self.screen.blit(self.sync, (0,0))

    def init(self):
        self.add_dtype('sync_square', bool, (1,))
        super().init()

    def _start_sync(self):
        self._tmp_fps = copy.deepcopy(self.fps)
        self.fps = self.sync_state_fps
        # if hasattr(self, 'decoder'):
        #     self.decoder.set_call_rate(1./self.fps)

    def _end_sync(self):
        self.fps = self._tmp_fps
        # if hasattr(self, 'decoder'):
        #     self.decoder.set_call_rate(1./self.fps)
        #     print("restore update rate")

    def _test_start_experiment(self, ts):
        return ts > self.sync_state_duration

    def _cycle(self):

        # Update the sync state
        if hasattr(self, 'sync_every_cycle') and self.sync_every_cycle:
            self.sync_state = not self.sync_state
        self.task_data['sync_square'] = copy.deepcopy(self.sync_state)
        
        # For OpenGL display, update the graphics
        if not hasattr(self, 'is_pygame_display'):
            color = self.sync_color_on if self.sync_state else self.sync_color_off
            self.sync_square.cube.color = color

        super()._cycle()

class CursorAnalogOut(traits.HasTraits):
    '''
    Output cursor x and z as analog voltages. Scales cursor position by 'cursor_out_gain', then
    centers on 1.15 volts. Outputs voltages using 12 bits of resolution on two channels of a 
    teensy microcontroller.
    '''

    cursor_gpio_port = traits.String("/dev/teensyao", desc="Port used for cursor analog out")
    cursor_x_pin = traits.Int(66, desc="Pin used to output cursor x position")
    cursor_z_pin = traits.Int(67, desc="Pin used to output cursor z position")
    cursor_x_ach = traits.Int(3, desc="Analog channel used to record cursor x position")
    cursor_z_ach = traits.Int(4, desc="Analog channel used to record cursor z position")
    cursor_out_gain = traits.Float(0.1, desc="Gain to control the output voltage of cursor position")
    hidden_traits = ["cursor_gpio_port", "cursor_x_pin", "cursor_z_pin", "cursor_x_ach", "cursor_z_ach", "cursor_out_gain"]

    def __init__(self, *args, **kwargs):

        self.cursor_output_gpio = TeensyGPIO(self.cursor_gpio_port)
        super().__init__(*args, **kwargs)

    def _cycle(self):
        pos = self.plant.get_endpoint_pos()
        voltage = pos*self.cursor_out_gain # pos (cm) * gain = voltage
        
        max_voltage = 3.3
        resolution = 12
        max_int = 2**resolution - 1
        float_values = (voltage + max_voltage/2)/max_voltage # (-1.15, +1.15) becomes (0, 1)
        float_values[float_values > 1] = 1.
        float_values[float_values < 0] = 0.
        if any(np.isnan(float_values)):
            super()._cycle()
            return
            
        int_values = (max_int*float_values).astype(int)
        self.cursor_output_gpio.analog_write(self.cursor_x_pin, int_values[0])
        # self.cursor_output_gpio.analog_write(self.cursor_y_pin, int_values[1])
        self.cursor_output_gpio.analog_write(self.cursor_z_pin, int_values[2])

        super()._cycle()



"""
Jacobian matrix operations.

@author: Peter Corke
@copyright: Peter Corke
"""

from numpy import *
from .utility import *
from .transform import *
from .kinematics import *
from numpy.linalg import norm
import numpy as np

def jacob0(robot, q):
    """
    Compute manipulator Jacobian in world coordinates for joint coordinates C{q}.

    The manipulator Jacobian matrix maps differential changes in joint space
    to differential Cartesian motion (world coord frame) of the end-effector

    M{dX = J dQ}

    @type robot: Robot
    @type q: vector M{n x 1}
    @param q: joint coordinate
    @rtype: matrix M{6 x n}
    @return: Manipulator Jacobian
    @see: L{jacobn}, L{diff2tr}, L{tr2diff}
    """


    q = np.mat(q).reshape(-1, 1)
    Jn = jacobn(robot, q) # Jacobian from joint to wrist space
    #   convert to Jacobian in base coordinates
    # Tn = fkine(robot,q) 
    Tn, _ = robot.fkine(q) # end-effector transformation
    R = t2r(Tn)
    return concatenate( ( concatenate( (R,zeros((3,3))) ,1) , concatenate( (zeros((3,3)),R) ,1) ))*Jn


def jacobn(robot, q):
    """
    Compute manipulator Jacobian in tool coordinates for joint coordinates C{q}.

    The manipulator Jacobian matrix maps differential changes in joint space
    to differential Cartesian motion (tool coord frame) of the end-effector.

    M{dX = J dQ}
		
    Reference
    =========
 	
 	Paul, Shimano, Mayer
    Differential Kinematic Control Equations for Simple Manipulators
    IEEE SMC 11(6) 1981
    pp. 456-460

    @type robot: Robot
    @type q: vector M{n x 1}
    @param q: joint coordinate
    @rtype: matrix M{6 x n}
    @return: Manipulator Jacobian
    @see: L{jacobn}, L{diff2tr}, L{tr2diff}
    """
    q = arg2array(q)
    n = robot.n
    L = robot.links
    J = mat([[],[],[],[],[],[]])
    U = robot.tool
    for j in range(n-1,-1,-1):
            if not robot.ismdh():    #standard DH convention
                U = L[j].tr(q[j])*U

            if L[j].sigma == 0: #revolute axis
                d = np.mat([[-U[0,0]*U[1,3] + U[1,0]*U[0,3]],
                            [-U[0,1]*U[1,3] + U[1,1]*U[0,3]],
                            [-U[0,2]*U[1,3] + U[1,2]*U[0,3]]])
                delta = U[2,0:3].T   # nz  oz  az
            else: #prismatic axis
                d = U[2,0:3].T       # nz  oz  az
                delta = np.zeros((3,1)) # 0   0   0

            J = concatenate((concatenate((d,delta)),J),1)
            
            if robot.ismdh(): #modified DH convention
                U=L[j].tr(q[j])*U
    return J



def tr2jac(t):
    """
    Compute a Jacobian to map differentials motion between frames.
    The M{6x6} Jacobian matrix to map differentials (joint velocity) between 
    frames related by the homogeneous transform C{t}.

    @rtype: matrix M{6 x 6}
    @return: Jacobian matrix
    @see: L{tr2diff}, L{diff2tr}
    """
    t = mat(t)
    return concatenate((
        concatenate((t[0:3,0].T, crossp(t[0:3,3],t[0:3,0]).T),1),
        concatenate((t[0:3,1].T, crossp(t[0:3,3],t[0:3,1]).T),1),
        concatenate((t[0:3,2].T, crossp(t[0:3,3],t[0:3,2]).T),1),
        concatenate((zeros((3,3)),t[0:3,0:3].T),1) ))



'''
Features for the phasephase motiontracker
'''
import numpy as np
from riglib import calibrations
from riglib.experiment import traits

mm_per_cm = 1./10

########################################################################################################
# Phasespace datasources
########################################################################################################
class MotionData(traits.HasTraits):
    '''
    Enable reading of raw motiontracker data from Phasespace system
    '''
    marker_count = traits.Int(8, desc="Number of markers to return")
    marker_num = traits.Int(1, desc="Which marker to use")

    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' sets up the DataSource for interacting with the 
        motion tracker system and registers the source with the SinkRegister so that the data gets saved to file as it is collected.
        '''
        from riglib import source
        src, mkw = self.source_class
        self.motiondata = source.DataSource(src, **mkw)
        from riglib import sink
        sink_manager = sink.SinkManager.get_instance()
        sink_manager.register(self.motiondata)
        super(MotionData, self).init()
    
    @property
    def source_class(self):
        '''
        Specify the source class as a function in case future descendant classes want to use a different type of source
        '''
        from riglib import motiontracker
        return motiontracker.make(self.marker_count), dict()

    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running. 
        See riglib.experiment.Experiment.run(). This 'run' method starts the motiontracker source prior to starting the experiment's 
        main thread/process, and handle any errors by stopping the source
        '''
        self.motiondata.start()
        try:
            super(MotionData, self).run()
        finally:
            self.motiondata.stop()
    
    def join(self):
        '''
        See riglib.experiment.Experiment.join(). Re-join the 'motiondata' source process before cleaning up the experiment thread
        '''
        self.motiondata.join()
        super(MotionData, self).join()
    
    def _start_None(self):
        '''
        Code to run before the 'None' state starts (i.e., the task stops)
        '''
        self.motiondata.stop()
        super(MotionData, self)._start_None()

    def _get_manual_position(self):
        ''' Sets the plant configuration based on motiontracker data. For manual control, uses
        motiontracker data. If no motiontracker data available, returns None'''

        #get data from motion tracker- take average of all data points since last poll
        pt = self.motiondata.get()
        if len(pt) > 0:
            pt = pt[:, self.marker_num, :]
            conds = pt[:, 3]
            inds = np.nonzero((conds>=0) & (conds!=4))[0]
            if len(inds) > 0:
                pt = pt[inds,:3]
                pt = pt.mean(0)
                pt = pt * mm_per_cm #self.convert_to_cm(pt)
            else: #if no usable data
                pt = None
        else: #if no new data
            pt = None

        return pt

class MotionSimulate(MotionData):
    '''
    Simulate presence of raw motiontracking system using a randomized spatial function
    '''
    @property
    def source_class(self):
        '''
        Specify the source class as a function in case future descendant classes want to use a different type of source
        '''        
        from riglib import motiontracker
        cls = motiontracker.make(self.marker_count, cls=motiontracker.Simulate)
        return cls, dict(radius=(100,100,50), offset=(-150,0,0))


class MotionAutoAlign(MotionData):
    '''Creates an auto-aligning motion tracker, for use with the 6-point alignment system'''
    autoalign = traits.Instance(calibrations.AutoAlign)
    
    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' adds a filter onto the motiondata source. See MotionData for further details.
        '''
        super(MotionAutoAlign, self).init()
        self.motiondata.filter = self.autoalign

    @property
    def source_class(self):
        '''
        Specify the source class as a function in case future descendant classes want to use a different type of source
        '''
        from riglib import motiontracker
        cls = motiontracker.make(self.marker_count, cls=motiontracker.AligningSystem)
        return cls, dict()



from numpy import *
import textwrap;
import sys;
import re;

# we use termios to pretty up the continuation prompting, only available under Unix
has_termios = False;
try:
    import termios;
    has_termios = True;
except:
    pass;
    
def parsedemo(s):
    """
    Helper to write demo files, based on minimum change from their Matlab format.
    The string contains
      - help text which is left justified and displayed directly to the screen
      - indented text is interpretted as a command, executed, and the output sent to
        screen
        
    The argument is Matlab code as per the demo files rtXXdemo.
    
    @note: Requires some minor changes to the demo file.
    @type s: string
    @param s: Demo string.
    """
    rstrip = re.compile(r'''\s*%.*$''')
    lines = s.split('\n');
    
    name = __file__;
    if has_termios:
        fd = sys.stdin.fileno()
        old = termios.tcgetattr(fd)
        new = termios.tcgetattr(fd)
        new[3] = new[3] & ~termios.ECHO          # lflags
    try:
        if has_termios:
            termios.tcsetattr(fd, termios.TCSADRAIN, new)
            
        text = '';
        for line in lines:
            if len(line) == 0:
                print;
            elif line[0] == '#':
                ## help text found, add it to the string
                text += line[1:].lstrip() + '\n';
                
                # a blank line means paragraph break
                if len(line) == 1 and text:
                    text = '';
            else:
                ## command encountered
                
                # flush the remaining help text
                if text:
                    text = '';
                
                cmd = line.strip();
                
                # special case, pause prompts the user to continue
                if cmd.startswith('pause'):
                    # prompt for continuation
                    sys.stderr.write('more? ')
                    raw_input();
                    
                    # remove the prompt from the screen
                    sys.stderr.write('\r        \r');
                    continue;
                
                # if it involves an assignment then we use exec else use eval.
                # we mimic the matlab behaviour in which a trailing semicolon inhibits
                # display of the result
                if cmd.startswith('from'):
                    exec(cmd)
                elif '=' in cmd:
                    e = cmd.split('=');
                    exec(rstrip.sub('', cmd))

                else:
                    result = eval(cmd.rstrip(';'));


    finally:
        if has_termios:
            termios.tcsetattr(fd, termios.TCSADRAIN, old)



import requests
import json

# This class is almost same as open ephys python tool other than timeout option
class OpenEphysHTTPServer:
    def __init__(self, address='10.155.205.108',timeout=0.5):
        self.address = 'http://' + address + ':37497'
        self.timeout = timeout

    def send(self, endpoint, payload=None):
        # Send a request to the server
        try:
            if payload is None:
                resp = requests.get(self.address + endpoint, timeout=self.timeout)
            else:
                resp = requests.put(self.address + endpoint, data = json.dumps(payload), timeout=self.timeout)
            response = resp.json()
        except Exception as e:
            print(e, 'Check if you started openephys in neuropixel computer')
            resp = {}
            resp['mode'] = 'No connection with openephys'
            response = resp
        return response

    def set_parent_dir(self,path):
        # Set the parent directory
        payload = {'parent_directory' : path}
        data = self.send('/api/recording', payload=payload)
        return data

    def set_prepend_text(self,text):
        # Set the prepend name for the recording directory
        payload = {'prepend_text' : text}
        data = self.send('/api/recording', payload=payload)
        return data

    def set_base_text(self,text):
        # Set the base name for the recording directory
        payload = {'base_text' : text}
        data = self.send('/api/recording', payload=payload)
        return data

    def set_append_text(self,text):
        # Set the append name for the recording directory
        payload = {'append_text' : text}
        data = self.send('/api/recording', payload=payload)
        return data

    def status(self):
        # Get the current status of the GUI (IDLE, ACQUIRE, or RECORD)
        return self.send('/api/status')['mode']

    def acquire(self):
        # Start acquisition
        payload = { 'mode' : 'ACQUIRE',}
        data = self.send('/api/status', payload=payload)
        return data['mode']

    def record(self):
        # Start recording data
        previous_mode = self.status()
        payload = {'mode' : 'RECORD',}
        data = self.send('/api/status', payload=payload)       
        return data['mode']

    def idle(self):
        # Stop acquisition or recording data
        previous_mode = self.status()
        payload = { 'mode' : 'IDLE',}
        data = self.send('/api/status', payload=payload)
        return data['mode']


# run with manage.py:
# > python manage.py shell
# > from db.migrate_db import *

from .boot_django import boot_django
boot_django()

from .tracker import models
import os
os.environ['DJANGO_SETTINGS_MODULE'] = 'db.settings'
import django
django.setup()

from db.tracker import models

'''
Update system paths
'''
#models.System.objects.filter(name='optitrack').update(path='/media/Optitrack')
#models.System.objects.filter(name='ecube').update(path='/media/NeuroAcq')

# optitrack_storage = "/media/Optitrack/"
# ecube_storage = "/media/NeuroAcq"

# for df in models.DataFile.objects.all():
#     if df.system.name == 'optitrack':
#         path = df.path
#         (path, filename) = os.path.split(path)
#         (path, session) = os.path.split(path)
#         (path, base_dir) = os.path.split(path)
#         df.path = os.path.join(optitrack_storage, base_dir, session, filename + '.tak')
#         print(df.path)
#         df.save()

#     if df.system.name == 'ecube':
#         path = df.path
#         (path, filename) = os.path.split(path)
#         df.path = os.path.join(ecube_storage, filename)
#         print(df.path)
#         df.save()

'''
Fix the sequence names to be reflections of their parameters
'''
# def make_seq_name(gen, params):
#     param_str = []
#     for (k, v) in params.items():
#         value_str = []
#         try:
#             for x in v['value']:
#                 value_str.append(str(x))
#         except:
#             value_str.append(str(v['value']))
#         param_str.append(f"{k}={','.join(value_str)}")
#     return f"{gen}:[{', '.join(param_str)}]"

# for seq in models.Sequence.objects.all():
#     json = seq.to_json()
#     params = json['params']
#     print(seq.name)
#     seq.name = make_seq_name(seq.generator.name, params)
#     print(seq.name)
#     print('-')
#     seq.save()

'''
Add metadata from xls files
'''
import pandas as pd
from datetime import datetime
df = pd.read_csv("/home/pagaiisland/Downloads/Beignet Records - Training 2021.csv", header=1)
handler_dict = {}
for index, row in df.iterrows():
    date = row["Date"]
    fdate = None
    if type(date) is str and date != '#REF!':
        try:
            fdate = datetime.strptime(date, "%m/%d/%Y")
        except:
            fdate = datetime.strptime(date+"/2021", "%m/%d/%Y")
    
    if fdate and not pd.isna(row["Handler"]):
        handler = str(row["Handler"]).lower()
        if "leo" in handler:
            handler = "Leo"
        elif "pavi" in handler:
            handler = "Pavi"
        elif "lydia" in handler:
            handler = "Lydia"
        elif "ryan" in handler:
            handler = "Ryan"
        else:
            print(handler)
            print(fdate)
            handler = "Leo"

        if fdate.date() not in handler_dict:
            handler_dict[fdate.date()] = dict.fromkeys([handler])
        else:
            handler_dict[fdate.date()].update(dict.fromkeys([handler]))

#print(handler_dict)

for te in models.TaskEntry.objects.all():
    date = te.date.date()
    handler = None
    project = "testing"
    session = ""
    if date in handler_dict and te.subject.name == "beignet":
        handlers = list(handler_dict[date])
        if len(handlers) == 1:
            handler = handlers[0]
        elif te.date.time() < datetime(year=2021,month=1,day=1,hour=13).time():
            handler = handlers[0]
            session = "morning"
        else:
            handler = handlers[1]
            session = "afternoon"

        if te.date.date() < datetime(year=2021, month=3, day=29).date():
            project = "open loop training"
        else:
            project = "manual control training"
    elif te.subject.name == "affi":
        handler = "Lydia"
        project = "open loop training"
    elif te.subject.name == "test":
        handler = "Leo"
    else:
        handler = "Lydia"
        
    experimenter = models.Experimenter.objects.get(name=handler)
    
    te.experimenter = experimenter
    te.project = project
    te.session = session
    te.save()



from riglib import experiment
from features.optitrack_features import Optitrack
from riglib.optitrack_client import optitrack
from datetime import datetime
import time
import numpy as np
import os
import natnet

import unittest


class TestOptitrack(unittest.TestCase):

    def test_client(self):
        optitrack_ip = '10.155.204.10'
        client = natnet.Client.connect(server=optitrack_ip)
        sys = optitrack.System(client)
        response = client.set_version(3,1)
        print(response)
        client._send_command_and_wait("LiveMode")
        time.sleep(1)
        print(sys.get())


    def test_datasource(self):
        optitrack_ip = '10.155.204.10'
        client = natnet.Client.connect(server=optitrack_ip)
        from riglib import source
        motiondata = source.DataSource(optitrack.make(optitrack.System, client, "rigid body", 1))
        motiondata.start()
        time.sleep(0.5)

        data = motiondata.get()
        motiondata.stop()

        # print(data)




if __name__ == '__main__':
    unittest.main()